# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import random
from typing import Tuple, Union, Dict, Any

from forte.common.configuration import Config
from forte.processors.data_augment.algorithms.single_annotation_op import (
    SingleAnnotationAugmentOp,
)
from forte.utils.utils import create_class_with_kwargs

__all__ = [
    "DictionaryReplacementOp",
]

from ft.onto.base_ontology import Token


class DictionaryReplacementOp(SingleAnnotationAugmentOp):
    r"""
    This class is a replacement op utilizing the dictionaries,
    such as WORDNET, to replace the input word with an synonym.
    Part-of-Speech(optional) can be provided to the wordnet for
    retrieving synonyms with the same POS. It will sample from a
    Bernoulli distribution to decide whether to replace the input,
    with `prob` as the probability of replacement.
    """

    def __init__(self, configs: Union[Config, Dict[str, Any]]):
        super().__init__(configs)
        self.dictionary = create_class_with_kwargs(
            self.configs["dictionary_class"], class_args={}
        )

    def single_annotation_augment(self, input_anno: Token) -> Tuple[bool, str]:  # type: ignore
        r"""
        This function replaces a word with synonyms from a WORDNET dictionary.

        Args:
            input_anno: The input word.
        Returns:
            A tuple of two values, where the first element is a boolean value
            indicating whether the replacement happens, and the second
            element is the replaced string.
        """
        # If the replacement does not happen, return False.
        if random.random() > self.configs.prob:
            return False, input_anno.text
        word = input_anno.text
        pos_tag = input_anno.pos
        lang = self.configs.lang
        synonyms = self.dictionary.get_synonyms(word, pos_tag, lang)
        if len(synonyms) == 0:
            return False, word
        return True, random.choice(synonyms)

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        r"""
        Returns:
            A dictionary with the default config for this processor.
        Following are the keys for this dictionary:
            - `dictionary` (dict):
                The full qualified name of the dictionary class.
            - `prob` (float):
                The probability of replacement, should fall in [0, 1].
                Default value is 0.1
            - `lang` (str):
                The language of the text.
        """

        dict_name = (
            "forte.processors.data_augment."
            "algorithms.dictionary.WordnetDictionary"
        )
        return {
            "dictionary_class": dict_name,
            "prob": 0.5,
            "lang": "eng",
        }



# Copyright 2021 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utility functions related to processors.
"""

__all__ = ["record_types_and_attributes_check", "collect_input_pack_record"]

from typing import Dict, Set
from forte.data.base_pack import PackType
from forte.common import ExpectedRecordNotFound
from forte.common.resources import Resources


def record_types_and_attributes_check(
    expectation: Dict[str, Set[str]], input_pack_record: Dict[str, Set[str]]
):
    r"""Check if any types or attributes in expectation dictionary doesn't
    match with input_pack.record. If not, an error of
    :class:`~forte.common.exception.ExpectedRecordNotFound` will be raised.

    Args:
        expectation: Dictionary of types and their attributes required for
            the current processor/evaluator.
        input_pack_record: The input pack record content combined with
            all the parent types and attributes collected from
            `merged_entry_tree`.

    Returns:

    """
    # pylint: disable=protected-access
    if expectation is not None:
        # check if expected types are in input pack.
        for expected_t in expectation:
            if expected_t not in input_pack_record.keys():
                raise ExpectedRecordNotFound(
                    f"The record type {expected_t} is not found in "
                    f"meta of the prediction datapack."
                )
            else:
                expected_value = expectation.get(expected_t)
                if expected_value is not None:
                    for expected_t_v in expected_value:
                        if expected_t_v not in input_pack_record.get(
                            expected_t, []
                        ):
                            raise ExpectedRecordNotFound(
                                f"The record attribute type "
                                f"{expected_t_v} is not found in "
                                f"attribute of record {expected_t} "
                                f"in meta of the input datapack."
                            )


def collect_input_pack_record(
    resources: Resources, input_pack: PackType
) -> Dict[str, Set[str]]:
    # pylint: disable=protected-access
    r"""Method to collect the type and attributes from the input pack and if
    :attr:`~forte.pipeline.Pipeline.resource` has `onto_specs` as key
    and ontology specification file path as value, then
    `merged_entry_tree` that has all the entries in ontology specification
    file would be populated. All the parent entry nodes of the input pack
    would be collected from this tree and add to the returned record
    dictionary for later comparison to enable subclass type checking.

    Args:
        resources: The pipeline attribute that stores and passes resources on
            the pipeline level.
        input_pack: The input datapack.

    Returns:
        input_pack_record: The input pack record content combined with
        all the parent types and attributes collected from
        merged_entry_tree

    """
    input_pack_record = input_pack._meta.record.copy()
    if resources.get("merged_entry_tree"):
        merged_entry_tree = resources.get("merged_entry_tree")
        merged_entry_tree.collect_parents(input_pack_record)
    return input_pack_record



# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This file implements BaseExtractor, which is the abstract class other
extractors will inherit from.
"""
import logging
from abc import ABC, abstractmethod
from typing import Tuple, List, Dict, Any
from typing import Union, Hashable, Iterable, Optional

from forte.common.configuration import Config
from forte.data.converter.feature import Feature
from forte.data.data_pack import DataPack
from forte.data.ontology import Annotation
from forte.data.vocabulary import Vocabulary

logger = logging.getLogger(__name__)

__all__ = ["BaseExtractor"]


class BaseExtractor(ABC):
    r"""The functionality of Extractor is as followed. Most of
    the time, a user will not need to call this class explicitly,
    they will be called by the framework.

        1. Build vocabulary.
        2. Extract feature from datapack.
        3. Perform pre-evaluation action on datapack.
        4. Add prediction to datapack.

    Explanation:

        - Vocabulary:
          Vocabulary is maintained as an attribute
          in extractor. It will store the mapping from element
          to index, which is an integer, and representation,
          which could be an index integer or one-hot vector
          depending on the configuration of the vocabulary.
          Check :class:`~forte.data.vocabulary.Vocabulary` for
          more details.
        - Feature:
          A feature basically wraps the data we want from
          one instance in a datapack. For example, the instance
          can be one sentence in a datapack. Then the data wrapped
          by the feature could be the token text of this sentence.
          The data is already converted as list of indexes using
          vocabulary. Besides the data, other information like the
          raw data before indexing and some meta_data will also be
          stored in the feature. Check
          :class:`~forte.data.converter.Feature` for more
          details.
        - Remove feature / Add prediction:
          Removing feature means remove
          the existing data in the datapack. If we remove the feature
          in the pack, then extracting feature will return empty list.
          Adding prediction means we add the prediction from model
          back to the datapack. If a datapack has some old data (for
          example, the golden data in the test set), we can first
          remove those data and then add our model prediction to
          the pack.

    Attributes:
        config: An instance of `Dict` or :class:`~forte.common.configuration.Config` that
            provides configurable options. See
            :meth:`~forte.data.base_extractor.BaseExtractor.default_configs`
            for available options and default values.

    """
    _VOCAB_ERROR_MSG = (
        "When vocab_method is raw, vocabulary "
        "will not be built. Functions operating "
        "on vocabulary should not be called."
    )

    def __init__(self):
        self._vocab: Optional[Vocabulary] = None
        self.config: Config = None
        self._vocab_method = None

    def initialize(self, config: Union[Dict, Config]):
        self.config = Config(config, self.default_configs())

        if self.config.vocab_method != "custom":
            self._vocab = Vocabulary(
                method=self.config.vocab_method,
                use_pad=self.config.need_pad,
                use_unk=self.config.vocab_use_unk,
                pad_value=self.config.pad_value,
                unk_value=self.config.unk_value,
            )
        else:
            self._vocab = None
        self._vocab_method = self.config.vocab_method

    @classmethod
    def default_configs(cls):
        r"""Returns a dictionary of default hyper-parameters.

        Here:

        - vocab_method (str)
          What type of vocabulary is used for this extractor. `custom`,
          `indexing`, `one-hot` are supported, default is `indexing`.
          Check the behavior of vocabulary under different setting
          in :class:`~forte.data.vocabulary.Vocabulary`

        - context_type (str): The fully qualified name of the context used to
          group the extracted features, for example, it could be a
          `ft.onto.base_ontology.Sentence`. If this is `None`, features from
          in the whole data pack will be grouped together. Default is None.
          This value could be mandatory for some processors, which will be
          documented and specified by the specific processor implementation.

        - vocab_use_unk (bool)
          Whether the `<UNK>` element should be added to vocabulary.
          Default is true.

        - need_pad (bool)
          Whether the `<PAD>` element should be added to vocabulary. And
          whether the feature need to be batched and padded. Default is True.

        - pad_value (int)
          A customized value/representation to be used for
          padding. This value is only needed when `use_pad` is True.
          Default is None, where the value of padding is determined by
          the system.

        - unk_value (int)
          A customized value/representation to be used for
          unknown value (`unk`). This value is only needed when
          `vocab_use_unk` is True. Default is None, where the value
          of `UNK` is determined by the system.

        """
        return {
            "vocab_method": "indexing",
            "context_type": None,
            "vocab_use_unk": True,
            "need_pad": True,
            "pad_value": None,
            "unk_value": None,
        }

    @property
    def vocab_method(self) -> str:
        return self._vocab_method

    @property
    def vocab(self) -> Optional[Vocabulary]:
        """
        Getter of the vocabulary class.

        Returns: The vocabulary. None if the vocabulary is not set.

        """
        return self._vocab

    @vocab.setter
    def vocab(self, vocab: Vocabulary):
        """
        Setter of the vocabulary, used when user build the vocabulary
        externally.

        Args:
            vocab: The vocabulary to be assigned.

        Returns:

        """
        self._vocab = vocab

    def get_pad_value(self) -> Union[None, int, List[int]]:
        if self.vocab is not None:
            return self.vocab.get_pad_value()
        else:
            return None

    def vocab_items(self) -> Iterable[Tuple[Hashable, int]]:
        if self.vocab is None:
            raise AttributeError(self._VOCAB_ERROR_MSG)
        return self.vocab.vocab_items()

    def add(self, element: Hashable):
        if self.vocab is None:
            raise AttributeError(self._VOCAB_ERROR_MSG)
        return self.vocab.add_element(element)

    def has_element(self, element: Hashable) -> bool:
        if self.vocab is None:
            raise AttributeError(self._VOCAB_ERROR_MSG)
        return self.vocab.has_element(element)

    def element2repr(self, element: Hashable) -> Union[int, List[int]]:
        if self.vocab is None:
            raise AttributeError(self._VOCAB_ERROR_MSG)
        return self.vocab.element2repr(element)

    def id2element(self, idx: int) -> Any:
        if self.vocab is None:
            raise AttributeError(self._VOCAB_ERROR_MSG)
        return self.vocab.id2element(idx)

    def predefined_vocab(self, predefined: Iterable):
        r"""Populate the vocabulary with predefined values. You can also extend
        this method to customize the ways to handle the vocabulary.

        Overwrite instruction:

            1. Take out elements from predefined.
            2. Make modification on the elements based on the need of the
               extractor.
            3. Use :meth:`add` function to add the element into vocabulary.

        Args:
            predefined (Iterable): A collections that contains the elements to
              be added into the vocabulary.
        """
        for element in predefined:
            self.add(element)

    def update_vocab(
        self, pack: DataPack, context: Optional[Annotation] = None
    ):
        r"""Populate the vocabulary needed by the extractor. This can be
        implemented by a specific extractor. The populated vocabulary can be
        used to map features/items to numeric representations. If you use a
        pre-specified vocabulary, you may not need to use this function.

        Overwrite instructions:

            1. Get all entries of the type of interest, such as all the
            `Token`s in the data pack.
            2. Use :meth:`~forte.data.vocabulary.Vocabulary.add` to add those
            element into `self._vocab`.

        Args:
            pack: The input data pack.
            context: The context is an Annotation entry where
                features will be extracted within its range. If None, then the
                whole data pack will be used as the context. Default is None.
        """
        pass

    @abstractmethod
    def extract(
        self, pack: DataPack, context: Optional[Annotation] = None
    ) -> Feature:
        """This method should be implemented to extract features from a
        datapack.

        Args:
            pack: The input data pack that contains the features.
            context: The context is an Annotation entry where
                features will be extracted within its range. If None, then the
                whole data pack will be used as the context. Default is None.

        Returns: Features inside this instance stored as a
        `~forte.data.converter.feature.Feature` instance.

        """
        raise NotImplementedError

    def pre_evaluation_action(
        self, pack: DataPack, context: Optional[Annotation]
    ):
        r"""This function is performed on the pack before the evaluation
        stage, allowing one to perform some actions before the evaluation.
        For example, you can remove entries or remove some attributes of
        the entry. By default, this function will not do anything.

        Args:
            pack: The datapack that contains the current instance.
            context: The context is an Annotation entry
                where data are extracted within its range. If None, then the
                whole data pack will be used as the context. Default is None.
        """
        pass

    def add_to_pack(
        self,
        pack: DataPack,
        predictions: Any,
        context: Optional[Annotation] = None,
    ):
        r"""Add prediction of a model (normally in the form of a tensor)
        back to the pack. This function should have knowledge of the structure
        of the `prediction` to correctly populate the data pack values.

        This function can be roughly considered as the reverse operation of
        :meth:`~forte.data.base_extractor.BaseExtractor.extract`.

        Overwrite instruction:

            1. Get all entries from one instance in the pack.

            2. Convert predictions into elements that needs to be assigned
               to entries. You can use :meth:`~forte.data.vocabulary.id2element`
               to convert integers in the prediction into element via the
               vocabulary maintained by the extractor.

            3. Add the element to corresponding entry based on the need.

        Args:
            pack: The datapack to add predictions back.
            predictions: This is the output of the model, the format of
              which will be determined by the predict function defined in the
              Predictor.
            context: The context is an Annotation
                entry where predictions will be added to. This has the same
                meaning with `context` as in
                :meth:`~forte.data.base_extractor.BaseExtractor.extract`.
                If None, then the whole data pack will be used as the
                context. Default is None.
        """
        pass



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from typing import Iterator, Any
from ft.onto.base_ontology import Token, Sentence, Document, EntityMention
from forte.data.data_pack import DataPack
from forte.data.data_utils_io import dataset_path_iterator
from forte.data.base_reader import PackReader

__all__ = ["CoNLL03Reader"]


class CoNLL03Reader(PackReader):
    r""":class:`CoNLL03Reader` is designed to read in the CoNLL03 dataset.

    The dataset is from the following paper:
    `Sang, Erik F., and Fien De Meulder. "Introduction to the CoNLL-2003
    shared task: Language-independent named entity recognition."
    arXiv preprint cs/0306050 (2003).`

    Data could be downloaded from
    https://deepai.org/dataset/conll-2003-english

    Data format:
    Data files contains one line "-DOCSTART- -X- -X- O" to represent the
    start of a document. After that, each line will contain one word and
    an empty line represent the start of a new sentence. Each line contains
    four fields, the word, its part-of-speech tag, its chunk tag and its
    named entity tag.

    Example:
        EU NNP B-NP B-ORG
        rejects VBZ B-VP O
        German JJ B-NP B-MISC
        call NN I-NP O
        to TO B-VP O
        boycott VB I-VP O
        British JJ B-NP B-MISC
        lamb NN I-NP O
        . . O O
    """

    def _collect(self, conll_directory) -> Iterator[Any]:  # type: ignore
        r"""Iterator over conll files in the data_source.

        Args:
            conll_directory: directory to the conll files.

        Returns: Iterator over files in the path with conll extensions.
        """
        logging.info("Reading .conll from %s", conll_directory)
        return dataset_path_iterator(conll_directory, "conll")

    def _cache_key_function(self, collection: str) -> str:
        return os.path.basename(collection)

    def _parse_pack(self, collection: str) -> Iterator[DataPack]:
        with open(collection, "r", encoding="utf8") as doc:
            pack_id: int = 0

            pack: DataPack = DataPack()
            text: str = ""
            offset: int = 0
            has_rows: bool = False

            sentence_begin: int = 0
            sentence_cnt: int = 0

            # NER tag is either "O" or in the format "X-Y",
            # where X is one of B, I,
            # Y is a tag like ORG, PER etc
            prev_y = None
            prev_x = None
            start_index = -1

            for line in doc:
                line = line.strip()

                if line.find("DOCSTART") != -1:
                    # Skip the first DOCSTART.
                    if offset == 0:
                        continue
                    # Add remaining sentence.
                    if has_rows:
                        # Add the last sentence if exists.
                        Sentence(pack, sentence_begin, offset - 1)
                        sentence_cnt += 1

                    pack.set_text(
                        text, replace_func=self.text_replace_operation
                    )
                    Document(pack, 0, len(text))
                    pack.pack_name = collection + "_%d" % pack_id
                    pack_id += 1
                    yield pack

                    # Create a new datapack.
                    pack = DataPack()
                    text = ""
                    offset = 0
                    has_rows = False

                    sentence_begin = 0
                    sentence_cnt = 0

                    prev_y = None
                    prev_x = None
                    start_index = -1

                elif line != "" and not line.startswith("#"):
                    conll_components = line.split()

                    word = conll_components[0]
                    pos = conll_components[1]
                    chunk_id = conll_components[2]

                    ner_tag = conll_components[3]

                    # A new ner tag occurs.
                    if ner_tag == "O" or ner_tag.split("-")[0] == "B":
                        # Add previous ner tag to sentence if it exists.
                        if prev_y is not None:
                            entity_mention = EntityMention(
                                pack, start_index, offset - 1
                            )
                            entity_mention.ner_type = prev_y

                        # Start process current ner tag.
                        if ner_tag == "O":
                            # Current ner tag is O, reset information.
                            prev_x = None
                            prev_y = None
                            start_index = -1
                        else:
                            # Current ner tag is B.
                            prev_x = "B"
                            prev_y = ner_tag.split("-")[1]
                            start_index = offset
                    # This ner tag is connected to previous one.
                    else:
                        x, y = ner_tag.split("-")
                        assert x == "I", "Unseen tag %s in the file." % x
                        assert y == prev_y, "Error in %s." % ner_tag
                        assert prev_x in ("B", "I"), "Error in %s." % ner_tag
                        prev_x = "I"

                    word_begin = offset
                    word_end = offset + len(word)

                    # Add tokens.
                    token = Token(pack, word_begin, word_end)
                    token.pos = pos
                    token.chunk = chunk_id

                    text += word + " "
                    offset = word_end + 1
                    has_rows = True
                else:
                    if not has_rows:
                        # Skip consecutive empty lines.
                        continue
                    # Add sentence
                    Sentence(pack, sentence_begin, offset - 1)

                    # Handle the last ner tag if exists.
                    if prev_x is not None:
                        entity_mention = EntityMention(
                            pack, start_index, offset - 1
                        )
                        entity_mention.ner_type = prev_y

                    # Reset information.
                    sentence_cnt += 1
                    has_rows = False
                    prev_y = None
                    prev_x = None
                    sentence_begin = offset

            if has_rows:
                # Add the last sentence if exists.
                Sentence(pack, sentence_begin, offset - 1)
                sentence_cnt += 1

            pack.set_text(text, replace_func=self.text_replace_operation)
            Document(pack, 0, len(text))
            pack.pack_name = os.path.basename(collection)

            yield pack



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from abc import abstractmethod
from typing import (
    Dict,
    List,
    Iterable,
    Union,
    Optional,
    Tuple,
    Type,
    Generic,
    Iterator,
    Any,
)

from forte.common import ValidationError, ProcessorConfigError
from forte.common.configurable import Configurable
from forte.common.configuration import Config
from forte.data.base_pack import PackType
from forte.data.converter import Feature
from forte.data.data_pack import DataPack
from forte.data.data_utils_io import merge_batches, batch_instances
from forte.data.multi_pack import MultiPack
from forte.data.ontology.core import EntryType
from forte.data.ontology.top import Annotation
from forte.utils import get_class

__all__ = [
    "ProcessingBatcher",
    "FixedSizeDataPackBatcherWithExtractor",
    "FixedSizeRequestDataPackBatcher",
    "FixedSizeMultiPackProcessingBatcher",
    "FixedSizeDataPackBatcher",
]


class ProcessingBatcher(Generic[PackType], Configurable):
    r"""This defines the basis interface of the batcher used in
    :class:`~forte.processors.base.batch_processor.BaseBatchProcessor`. This
    Batcher only batches data sequentially. It receives new packs dynamically
    and cache the current packs so that the processors can pack prediction
    results into the data packs.
    """

    def __init__(self):
        super().__init__()
        self.current_batch: Dict = {}
        self.data_pack_pool: List[PackType] = []
        self.current_batch_sources: List[int] = []

        self._cross_pack: bool = True
        self.configs: Config = Config({}, {})

    def initialize(self, config: Union[Config, Dict]):
        r"""The implementation should initialize the batcher and setup the
        internal states of this batcher. This function will be called at the
        pipeline initialize stage.

        Returns:
            None
        """
        self.configs = self.make_configs(config)
        self._cross_pack = self.configs.cross_pack
        self.current_batch.clear()
        self.data_pack_pool.clear()
        self.current_batch_sources.clear()

    @abstractmethod
    def _should_yield(self) -> bool:
        r"""User should implement this based on the state of the batcher to
        indicate whether the batch criteria is met and the batcher should yield
        the current batch. For example, whether the number of instances reaches
        the batch size.

        Returns:
            None
        """
        raise NotImplementedError

    def flush(
        self,
    ) -> Iterator[Tuple[List[PackType], List[Optional[Annotation]], Dict]]:
        r"""Flush the remaining data.

        Returns:
            A triplet contains datapack, context instance and batched data.

            .. note::

                For backward compatibility issues, this function
                return list of None contexts.
        """
        if self.current_batch:
            yield (
                self.data_pack_pool,
                [None] * len(self.data_pack_pool),
                self.current_batch,
            )
            self.current_batch = {}
            self.current_batch_sources = []
            self.data_pack_pool = []

    def get_batch(
        self, input_pack: PackType
    ) -> Iterator[Tuple[List[PackType], List[Optional[Annotation]], Dict]]:
        r"""By feeding data pack to this function, formatted features will
        be yielded based on the batching logic. Each element in the iterator is
        a triplet of datapack, context instance and batched data.

        Args:
            input_pack: The input data pack to get features from.

        Returns:
            An iterator of A tuple contains datapack, context instance and
            batch data.

            .. note::

                For backward compatibility issues, this function
                return a list of `None` as contexts.
        """
        batch_count = 0

        # cache the new pack and generate batches
        for (data_batch, instance_num) in self._get_data_batch(input_pack):
            self.current_batch = merge_batches([self.current_batch, data_batch])
            self.current_batch_sources.append(instance_num)
            self.data_pack_pool.extend([input_pack] * instance_num)

            # Yield a batch on two conditions.
            # 1. If we do not want to have batches from different pack, we
            # should yield since this pack is exhausted.
            # 2. We should also yield when the batcher condition is met:
            # i.e. ``_should_yield()`` is True.
            if not self._cross_pack or self._should_yield():
                batch_count += 1
                yield (
                    self.data_pack_pool,
                    [None] * len(self.data_pack_pool),
                    self.current_batch,
                )
                self.current_batch = {}
                self.current_batch_sources = []
                self.data_pack_pool = []

    def _get_data_batch(
        self,
        data_pack: PackType,
    ) -> Iterable[Tuple[Dict, int]]:
        r"""The abstract function that a batcher need to implement, to collect
        data from input data packs. It should yield data in the format of a
        tuple that contains the actual data points and the number of data
        points.

        These data points will be collected and organized in batches
        by the batcher, and can be obtained from the `get_batch` method.

        Args:
            data_pack: The data pack to retrieve data from.

        Returns:
            An iterator of tuples ``(batch, cnt)``, ``batch`` is a dict
            containing the required annotations and context, and ``cnt`` is
            the number of instances in the batch.
        """
        raise NotImplementedError

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        """
        Define the basic configuration of a batcher. Implementation of the
        batcher can extend this function to include more configurable
        parameters but need to keep the existing ones defined in this base
        class.

        Here, the available parameters are:

          - `use_coverage_index`: A boolean value indicates whether the
            batcher will try to build the coverage index based on the data
            request. Default is True.

          - `cross_pack`: A boolean value indicates whether the batcher can
            go across the boundary of data packs when there is no enough data
            to fill the batch.


        Returns:
            The default configuration.

        """
        return {
            "use_coverage_index": True,
            "cross_pack": True,
        }


# TODO: shouldn't implement a special extractor because we use extractors.
class FixedSizeDataPackBatcherWithExtractor(ProcessingBatcher):
    r"""This batcher uses extractor to extract features from
    dataset and group them into batch. In this class, more pools
    are added. One is `instance_pool`, which is used to record the
    instance from which feature is extracted. The other one is
    `feature_pool`, which is used to record features before they
    can be yield in batch.
    """

    def __init__(self):
        super().__init__()
        self._context_type: Type[EntryType] = None
        self._feature_scheme: Dict = {}
        self.batch_size: int = -1

        self.instance_pool: List[Annotation] = []
        self.feature_pool: List[Dict[str, Feature]] = []
        self.pool_size = 0
        self.batch_is_full = False

    def initialize(self, config: Config):
        super().initialize(config)

        if config["context_type"] is None:
            raise AttributeError("'context_type' cannot be None.")

        if config["batch_size"] is None:
            raise AttributeError("'batch_size' cannot be None.")

        if isinstance(config["context_type"], str):
            self._context_type = get_class(config["context_type"])
        else:
            self._context_type = config["context_type"]

        if not issubclass(self._context_type, Annotation):
            raise ValidationError(
                f"The provided context type {self._context_type} "
                f"is not an Annotation type."
            )

        self.batch_size = config["batch_size"]

        self.instance_pool.clear()
        self.feature_pool.clear()
        self.pool_size = 0
        self.batch_is_full = False

    def add_feature_scheme(self, tag: str, scheme: Dict[str, Any]):
        """
        Add feature scheme to the batcher.

        Args:
            tag (str): The name/tag of the scheme.
            scheme (str): The scheme content, which should be a dict
              containing the extractor and converter used to create features.
        """
        self._feature_scheme[tag] = scheme

    def collate(
        self, features_collection: List[Dict[str, Feature]]
    ) -> Dict[str, Dict[str, Any]]:
        r"""This function use the :class:`~forte.data.converter.converter.Converter`
        interface to turn a list of features into batches, where each feature
        is converted to tensor/matrix format. The resulting features are
        organized as a dictionary, where the keys are the feature names/tags,
        and the values are the converted features. Each feature contains the
        data and mask in `MatrixLike` form, as well as the original raw
        features.

        Args:
            features_collection: A list of features.

        Returns:
            A instance of `Dict[str, Union[Tensor, Dict]]`, which
            is a batch of features.
        """
        collections: Dict[str, List[Feature]] = {}
        for features in features_collection:
            for tag, feat in features.items():
                if tag not in collections:
                    collections[tag] = []
                collections[tag].append(feat)

        converted: Dict[str, Dict[str, Any]] = {}
        for tag, features in collections.items():
            converter = self._feature_scheme[tag]["converter"]
            data, masks = converter.convert(features)
            converted[tag] = {
                "data": data,
                "masks": masks,
                "features": features,
            }
        return converted

    def _should_yield(self) -> bool:
        return self.batch_is_full

    def flush(
        self,
    ) -> Iterator[
        Tuple[
            List[PackType],
            List[Optional[Annotation]],
            Dict[str, Dict[str, Any]],
        ]
    ]:
        r"""Flush data in batches. Each return value contains a tuple of 3
        items: the corresponding data pack, the list of annotation objects
        that represent the context type, and the features."""
        if self.pool_size > 0:
            yield (  # type: ignore
                self.data_pack_pool,
                self.instance_pool,
                self.collate(self.feature_pool),
            )
            self.data_pack_pool = []
            self.instance_pool = []
            self.feature_pool = []
            self.pool_size = 0

    def get_batch(
        self,
        input_pack: PackType,
    ) -> Iterator[Tuple[List[PackType], List[Optional[Annotation]], Dict]]:
        r"""By feeding data pack to this function, formatted features will
        be yielded based on the batching logic. Each element in the iterator is
        a triplet of datapack, context instance and batched data.

        Args:
            input_pack: The input data pack to get features from.

        Returns:
            An iterator of a tuple contains datapack, context instance and
            batch data.
        """
        # cache the new pack and generate batches

        for (batch, num) in self._get_data_batch(input_pack):
            self.data_pack_pool.extend(batch["packs"])
            self.instance_pool.extend(batch["contexts"])
            self.feature_pool.extend(batch["features"])
            self.pool_size += num

            # Yield a batch on two conditions.
            # 1. If we do not want to have batches from different pack, we
            # should yield since this pack is exhausted.
            # 2. We should also yield when the batcher condition is met:
            # i.e. ``_should_yield()`` is True.
            if not self._cross_pack or self._should_yield():
                yield from self.flush()  # type: ignore

    def _get_data_batch(
        self,
        data_pack: PackType,
    ) -> Iterable[Tuple[Dict[Any, Any], int]]:
        r"""Get data batches based on the requests.

        Args:
            data_pack: The data pack to retrieve data from.
        """
        packs: List[PackType] = []
        contexts: List[Annotation] = []
        features_collection: List[Dict[str, Feature]] = []
        current_size = self.pool_size

        for instance in data_pack.get(self._context_type):
            contexts.append(instance)
            features = {}
            for tag, scheme in self._feature_scheme.items():
                features[tag] = scheme["extractor"].extract(data_pack)
            packs.append(data_pack)
            features_collection.append(features)

            if len(contexts) == self.batch_size - current_size:
                self.batch_is_full = True

                batch = {
                    "packs": packs,
                    "contexts": contexts,
                    "features": features_collection,
                }

                yield batch, len(contexts)
                self.batch_is_full = False
                packs = []
                contexts = []
                features_collection = []
                current_size = self.pool_size

        # Flush the remaining data.
        if len(contexts) > 0:
            batch = {
                "packs": packs,
                "contexts": contexts,
                "features": features_collection,
            }
            yield batch, len(contexts)

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        """
        Defines the configuration of this batcher, here:

            - context_type: The context scope to extract data from. It could be
              a annotation class or a string that is the fully qualified name
              of the annotation class.

            - feature_scheme: A dictionary of (extractor name, extractor) that
              can be used to extract features.

            - batch_size: The batch size, default is 10.

        Returns:
            The default configuration structure.

        """
        return {
            "context_type": None,
            "feature_scheme": None,
            "batch_size": 10,
        }


class FixedSizeDataPackBatcher(ProcessingBatcher[DataPack]):
    def __init__(self):
        super().__init__()
        self.batch_is_full = False

    def initialize(self, config: Config):
        super().initialize(config)
        self.batch_is_full = False

    def _should_yield(self) -> bool:
        return self.batch_is_full

    @abstractmethod
    def _get_instance(self, data_pack: DataPack) -> Iterator[Dict[str, Any]]:
        """
        Get instance from the data pack. By default, this function will use
        the `requests` in configuration to get data. One can implement this
        function to extract data instance.

        Args:
            data_pack: The data pack to extract data from.

        Returns:
            None
        """
        raise NotImplementedError

    def _get_data_batch(
        self,
        data_pack: DataPack,
    ) -> Iterable[Tuple[Dict, int]]:
        r"""Get batches from a dataset  with ``batch_size``, It will yield data
        in the format of a tuple that contains the actual data points and the
        number of data points.

        The data points are generated by querying the data pack using the
        `context_type` and `requests` configuration via calling the
        :meth:`~forte.data.DataPack.get_data` method. Here, Each data point is
        in the same format returned by the `get_data` method, and the meaning
        of `context_type` and `requests` are exactly the same as the `get_data`
        method.

        Args:
            data_pack: The data pack to retrieve data from.

        Returns:
            An iterator of tuples ``(batch, cnt)``, ``batch`` is a dict
            containing the required entries and context, and ``cnt`` is
            the number of instances in the batch.
        """
        instances: List[Dict] = []
        current_size = sum(self.current_batch_sources)

        for data in self._get_instance(data_pack):
            instances.append(data)
            if len(instances) == self.configs.batch_size - current_size:
                batch = batch_instances(instances)
                self.batch_is_full = True
                yield batch, len(instances)
                instances = []
                self.batch_is_full = False

        # Flush the remaining data.
        if len(instances) > 0:
            batch = batch_instances(instances)
            yield batch, len(instances)

    @classmethod
    def default_configs(cls) -> Dict:
        """
        The configuration of a batcher.

        Here:

            - batch_size: the batch size, default is 10.


        Returns:
            The default configuration structure and default value.
        """
        return {
            "batch_size": 10,
        }


class FixedSizeRequestDataPackBatcher(FixedSizeDataPackBatcher):
    def initialize(self, config: Config):
        super().initialize(config)
        if self.configs.context_type is None:
            raise ProcessorConfigError(
                f"The 'context_type' config of {self.__class__.__name__} "
                f"cannot be None."
            )

    def _get_instance(self, data_pack: DataPack) -> Iterator[Dict[str, Any]]:
        """
        Get instance from the data pack. By default, this function will use
        the `requests` in configuration to get data. One can implement this
        function to extract data instance.

        Args:
            data_pack: The data pack to extract data from.

        Returns:
            None
        """
        yield from data_pack.get_data(
            self.configs.context_type, self.configs.requests.todict()
        )

    @classmethod
    def default_configs(cls) -> Dict:
        """
        The configuration of a batcher.

        Here:

            - context_type (str): The fully qualified name of an `Annotation`
              type, which will be used as the context to retrieve data from.
              For
              example, if a `ft.onto.Sentence` type is provided, then it will
              extract data within each sentence.
            - requests: The request detail. See
              :meth:`~forte.data.data_pack.DataPack.get_data` on what a request
              looks like.


        Returns:
            The default configuration structure and default value.
        """
        return {
            "context_type": None,
            "requests": {},
            "@no_typecheck": "requests",
        }


class FixedSizeMultiPackProcessingBatcher(ProcessingBatcher[MultiPack]):
    r"""A Batcher used in :class:`~forte.processors.base.batch_processor.MultiPackBatchProcessor`.

    .. note::

        this implementation is not finished.

    The Batcher calls the ProcessingBatcher inherently on each specified
    data pack in the MultiPack.

    It's flexible to query MultiPack so we delegate the task to the subclasses
    such as:

        - query all packs with the same ``context`` and ``input_info``.
        - query different packs with different ``context`` and
          ``input_info``.

    Since the batcher will save the data_pack_pool on the fly, it's not trivial
    to do batching and slicing multiple data packs in the same time
    """

    def __init__(self):
        super().__init__()
        self.batch_is_full = False
        self.input_pack_name: str = ""
        self.batch_size = -1

    def initialize(self, config: Config):
        super().initialize(config)
        self.batch_is_full = False

    def _should_yield(self) -> bool:
        return self.batch_is_full

    def _get_data_batch(
        self, multi_pack: MultiPack
    ) -> Iterable[Tuple[Dict, int]]:
        # TODO: Principled way of get data from multi pack?
        raise NotImplementedError

    @classmethod
    def default_configs(cls) -> Dict:
        return {"batch_size": 10}



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The reader that reads Open-IE extractions data into data pack.
Format:
https://github.com/gabrielStanovsky/oie-benchmark/tree/master/oie_corpus
"""
import logging
import os
from typing import Iterator, List

from forte.common.configuration import Config
from forte.common.exception import ProcessorConfigError
from forte.common.resources import Resources
from forte.data.data_pack import DataPack
from forte.data.data_utils_io import dataset_path_iterator
from forte.data.base_reader import PackReader
from ft.onto.base_ontology import Sentence, RelationLink, EntityMention

__all__ = ["OpenIEReader"]


class OpenIEReader(PackReader):
    r""":class:`OpenIEReader` is designed to read in the Open IE dataset used
    by Open Information Extraction task. The related paper can be found
    `here
    <https://gabrielstanovsky.github.io/assets/papers/emnlp16a/paper.pdf>`__.
    The related source code for generating this dataset can be found
    `here
    <https://github.com/gabrielStanovsky/oie-benchmark>`__.
    To use this Reader, you must follow the dataset format. Each line in
    the dataset should contain following fields:

    .. code-block:: none

        <sentence>\t<predicate_head>\t<full_predicate>\t<arg1>\t<arg2>....

    You can also find the dataset format `here
    <https://github.com/gabrielStanovsky/oie-benchmark/tree/master/oie_corpus>`__.
    """

    def initialize(self, resources: Resources, configs: Config):
        super().initialize(resources, configs)

        if configs.oie_file_extension is None:
            raise ProcessorConfigError(
                "Configuration oie_file_extension not provided."
            )

    def _collect(self, *args, **kwargs) -> Iterator[str]:
        # pylint: disable = unused-argument
        r"""Should be called with param ``oie_directory`` which is a path to a
        folder containing json files.

        Args:
            args: args[0] is the directory to the open ie files.
            kwargs:

        Returns: Iterator over files in the path with oie extensions.
        """
        oie_directory: str = args[0]
        oie_file_extension: str = self.configs.oie_file_extension
        logging.info(
            "Reading dataset from %s with extension %s",
            oie_directory,
            oie_file_extension,
        )
        return dataset_path_iterator(oie_directory, oie_file_extension)

    def _cache_key_function(self, oie_file: str) -> str:
        return os.path.basename(oie_file)

    def _parse_pack(self, file_path: str) -> Iterator[DataPack]:
        pack: DataPack = DataPack()
        text: str = ""
        offset: int = 0

        with open(file_path, "r", encoding="utf8") as f:
            for line in f:
                line = line.strip()
                if line != "":
                    oie_component: List[str] = line.split("\t")

                    # Add sentence.
                    sentence = oie_component[0]
                    text += sentence + "\n"
                    Sentence(pack, offset, offset + len(sentence))

                    # Find argument 1.
                    arg1_begin = sentence.find(oie_component[3]) + offset
                    arg1_end = arg1_begin + len(oie_component[3])
                    arg1: EntityMention = EntityMention(
                        pack, arg1_begin, arg1_end
                    )

                    # Find argument 2.
                    arg2_begin = sentence.find(oie_component[4]) + offset
                    arg2_end = arg2_begin + len(oie_component[4])
                    arg2: EntityMention = EntityMention(
                        pack, arg2_begin, arg2_end
                    )

                    head_relation = RelationLink(pack, arg1, arg2)
                    head_relation.rel_type = oie_component[2]

                    offset += len(sentence) + 1

        self.set_text(pack, text)
        pack.pack_name = os.path.basename(file_path)
        yield pack

    @classmethod
    def default_configs(cls):
        config: dict = super().default_configs()

        # Add OIE dataset file extension. The default is '.oie'
        config.update({"oie_file_extension": "oie"})
        return config



# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Class for back translation op. The input is translated
to another language, then translated back to the original language.
"""
import random
from typing import Any, Dict, Tuple

from forte.data.ontology import Annotation
from forte.processors.data_augment.algorithms.single_annotation_op import (
    SingleAnnotationAugmentOp,
)
from forte.common.configuration import Config
from forte.utils.utils import create_class_with_kwargs

__all__ = [
    "BackTranslationOp",
]


class BackTranslationOp(SingleAnnotationAugmentOp):
    r"""
    This class is a replacement op using back translation
    to generate data with the same semantic meanings. The
    input is translated to another language, then translated
    back to the original language, with pretrained
    machine-translation models.

    It will sample from a Bernoulli distribution to decide
    whether to replace the input, with `prob` as the probability
    of replacement.
    """

    def __init__(self, configs: Config):
        super().__init__(configs)

        self._validate_configs(configs)

        self.model_to = create_class_with_kwargs(
            configs["model_to"],
            class_args={
                "src_lang": configs["src_language"],
                "tgt_lang": configs["tgt_language"],
                "device": configs["device"],
            },
        )

        self.model_back = create_class_with_kwargs(
            configs["model_back"],
            class_args={
                "src_lang": configs["tgt_language"],
                "tgt_lang": configs["src_language"],
                "device": configs["device"],
            },
        )

    def _validate_configs(self, configs):
        prob = configs["prob"]
        if not prob or prob < 0 or prob > 1:
            raise ValueError("The prob should be a float between 0 and 1!")

        src_lang = configs["src_language"]
        if not src_lang or len(src_lang) == 0:
            raise ValueError("Please provide a valid source language!")

        tgt_lang = configs["tgt_language"]
        if not tgt_lang or len(tgt_lang) == 0:
            raise ValueError("Please provide a valid target language!")

        model_to = configs["model_to"]
        if not model_to or len(model_to) == 0:
            raise ValueError("Please provide a valid to-model!")

        model_back = configs["model_back"]
        if not model_back or len(model_back) == 0:
            raise ValueError("Please provide a valid back-model!")

        device = configs["device"]
        if device not in ("cpu", "cuda"):
            raise ValueError("The device must be 'cpu' or 'cuda'!")

    def single_annotation_augment(
        self, input_anno: Annotation
    ) -> Tuple[bool, str]:
        r"""
        This function replaces a piece of text with back translation.

        Args:
            input_anno: An annotation, could be a word, sentence
              or document.

        Returns:
            A tuple, where the first element is a boolean value indicating
            whether the replacement happens, and the second element is the
            replaced string.
        """
        # If the replacement does not happen, return False.
        if random.random() > self.configs["prob"]:
            return False, input_anno.text
        intermediate_text: str = self.model_to.translate(input_anno.text)
        return True, self.model_back.translate(intermediate_text)

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        """
        Returns:
            A dictionary with the default config for this processor.
        Following are the keys for this dictionary:

            - `augment_entry` (str):
                This indicates the entity that needs to
                be augmented. By default, this value is set to
                `ft.onto.base_ontology.Sentence`.

            - `prob` (float):
                The probability of replacement, should fall in [0, 1].
                The Default value is 0.5

            - `src_language` (str):
                The source language of back translation.

            - `tgt_language` (str):
                The target language of back translation.

            - `model_to` (str):
                The full qualified name of the model from
                source language to target language.

            - `model_back` (str):
                The full qualified name of the model from
                target language to source language.

            - `device` (str):
                "cpu" for the CPU or "cuda" for GPU. The Default
                value is cpu.
        """
        model_class_name = (
            "forte.processors.data_augment.algorithms."
            "machine_translator.MarianMachineTranslator"
        )
        return {
            "augment_entry": "ft.onto.base_ontology.Sentence",
            "prob": 0.5,
            "model_to": model_class_name,
            "model_back": model_class_name,
            "src_language": "en",
            "tgt_language": "fr",
            "device": "cpu",
        }



# DataStore constants
# The index storing begin location in the internal entry data of DataStore.
BEGIN_INDEX = 0

# The index storing end location in the internal entry data of DataStore.
END_INDEX = 1

# The index storing tid in the internal entry data of DataStore.
TID_INDEX = 2

# The index storing entry type in the internal entry data of DataStore.
ENTRY_TYPE_INDEX = 3

# The index storing entry index (specific to Link and Group type) in the internal
# entry data of DataStore.
ENTRY_INDEX_INDEX = -1

# The index where the first attribute appears in the internal entry data of DataStore.
ATTR_BEGIN_INDEX = 4



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from dataclasses import dataclass
from functools import total_ordering
from typing import Optional, Set, Tuple, Type, Any, Dict, Union, Iterable, List

import numpy as np

from forte.data.base_pack import PackType
from forte.data.ontology.core import (
    Entry,
    BaseLink,
    BaseGroup,
    MultiEntry,
    EntryType,
)
from forte.data.span import Span

__all__ = [
    "Generics",
    "Annotation",
    "Group",
    "Link",
    "MultiPackGeneric",
    "MultiPackGroup",
    "MultiPackLink",
    "Query",
    "SinglePackEntries",
    "MultiPackEntries",
    "AudioAnnotation",
]

QueryType = Union[Dict[str, Any], np.ndarray]


class Generics(Entry):
    def __init__(self, pack: PackType):
        super().__init__(pack=pack)


@total_ordering
class Annotation(Entry):
    r"""Annotation type entries, such as "token", "entity mention" and
    "sentence". Each annotation has a :class:`~forte.data.span.Span` corresponding to its offset
    in the text.

    Args:
        pack: The container that this annotation
            will be added to.
        begin: The offset of the first character in the annotation.
        end: The offset of the last character in the annotation + 1.
    """

    def __init__(self, pack: PackType, begin: int, end: int):
        self._span: Optional[Span] = None
        self._begin: int = begin
        self._end: int = end
        super().__init__(pack)

    def __getstate__(self):
        r"""For serializing Annotation, we should create Span annotations for
        compatibility purposes.
        """
        self._span = Span(self._begin, self._end)
        state = super().__getstate__()
        state.pop("_begin")
        state.pop("_end")
        return state

    def __setstate__(self, state):
        """
        For de-serializing Annotation, we load the begin, end from Span, for
        compatibility purposes.
        """
        super().__setstate__(state)
        self._begin = self._span.begin
        self._end = self._span.end

    @property
    def span(self) -> Span:
        # Delay span creation at usage.
        if self._span is None:
            self._span = Span(self._begin, self._end)
        return self._span

    @property
    def begin(self):
        return self._begin

    @property
    def end(self):
        return self._end

    def __eq__(self, other):
        r"""The eq function of :class:`Annotation`.
        By default, :class:`Annotation` objects are regarded as the same if
        they have the same type, span, and are generated by the same component.

        Users can define their own eq function by themselves but this must
        be consistent to :meth:`hash`.
        """
        if other is None:
            return False
        return (type(self), self.begin, self.end) == (
            type(other),
            other.begin,
            other.end,
        )

    def __lt__(self, other):
        r"""To support total_ordering, `Annotation` must implement
        `__lt__`. The ordering is defined in the following way:

        1. If the begin of the annotations are different, the one with larger
           begin will be larger.
        2. In the case where the begins are the same, the one with larger
           end will be larger.
        3. In the case where both offsets are the same, we break the tie using
           the normal sorting of the class name.
        """
        if self.begin == other.begin:
            if self.end == other.end:
                return str(type(self)) < str(type(other))
            return self.end < other.end
        else:
            return self.begin < other.begin

    @property
    def text(self):
        if self.pack is None:
            raise ValueError(
                "Cannot get text because annotation is not "
                "attached to any data pack."
            )
        return self.pack.get_span_text(self.begin, self.end)

    @property
    def index_key(self) -> int:
        return self.tid

    def get(
        self,
        entry_type: Union[str, Type[EntryType]],
        components: Optional[Union[str, Iterable[str]]] = None,
        include_sub_type=True,
    ) -> Iterable[EntryType]:
        """
        This function wraps the :meth:`~forte.data.data_pack.DataPack.get` method to find
        entries "covered" by this annotation. See that method for more information.

        Example:

            .. code-block:: python

                # Iterate through all the sentences in the pack.
                for sentence in input_pack.get(Sentence):
                    # Take all tokens from each sentence created by NLTKTokenizer.
                    token_entries = sentence.get(
                        entry_type=Token,
                        component='NLTKTokenizer')
                    ...

            In the above code snippet, we get entries of type
            :class:`~ft.onto.base_ontology.Token` within
            each ``sentence`` which were generated by ``NLTKTokenizer``. You
            can consider build coverage index between `Token` and `Sentence`
            if this snippet is frequently used.

        Args:
            entry_type: The type of entries requested.
            components: The component (creator)
                generating the entries requested. If `None`, will return valid
                entries generated by any component.
            include_sub_type: whether to consider the sub types of
                the provided entry type. Default `True`.

        Yields:
            Each `Entry` found using this method.

        """
        yield from self.pack.get(entry_type, self, components, include_sub_type)


class Link(BaseLink):
    r"""Link type entries, such as "predicate link". Each link has a parent node
    and a child node.

    Args:
        pack: The container that this annotation
            will be added to.
        parent: the parent entry of the link.
        child: the child entry of the link.
    """
    # this type Any is needed since subclasses of this class will have new types
    ParentType: Any = Entry
    ChildType: Any = Entry

    def __init__(
        self,
        pack: PackType,
        parent: Optional[Entry] = None,
        child: Optional[Entry] = None,
    ):
        self._parent: Optional[int] = None
        self._child: Optional[int] = None
        super().__init__(pack, parent, child)

    # TODO: Can we get better type hint here?
    def set_parent(self, parent: Entry):
        r"""This will set the `parent` of the current instance with given Entry
        The parent is saved internally by its pack specific index key.

        Args:
            parent: The parent entry.
        """
        if not isinstance(parent, self.ParentType):
            raise TypeError(
                f"The parent of {type(self)} should be an "
                f"instance of {self.ParentType}, but get {type(parent)}"
            )
        self._parent = parent.tid

    def set_child(self, child: Entry):
        r"""This will set the `child` of the current instance with given Entry.
        The child is saved internally by its pack specific index key.

        Args:
            child: The child entry.
        """
        if not isinstance(child, self.ChildType):
            raise TypeError(
                f"The parent of {type(self)} should be an "
                f"instance of {self.ParentType}, but get {type(child)}"
            )
        self._child = child.tid

    @property
    def parent(self):
        r"""Get ``tid`` of the parent node. To get the object of the parent
        node, call :meth:`get_parent`.
        """
        return self._parent

    @property
    def child(self):
        r"""Get ``tid`` of the child node. To get the object of the child node,
        call :meth:`get_child`.
        """
        return self._child

    def get_parent(self) -> Entry:
        r"""Get the parent entry of the link.

        Returns:
             An instance of :class:`~forte.data.ontology.core.Entry` that is the parent of the link.
        """
        if self.pack is None:
            raise ValueError(
                "Cannot get parent because link is not "
                "attached to any data pack."
            )
        if self._parent is None:
            raise ValueError("The parent of this entry is not set.")
        return self.pack.get_entry(self._parent)

    def get_child(self) -> Entry:
        r"""Get the child entry of the link.

        Returns:
             An instance of :class:`~forte.data.ontology.core.Entry` that is the child of the link.
        """
        if self.pack is None:
            raise ValueError(
                "Cannot get child because link is not"
                " attached to any data pack."
            )
        if self._child is None:
            raise ValueError("The child of this entry is not set.")
        return self.pack.get_entry(self._child)


# pylint: disable=duplicate-bases
class Group(BaseGroup[Entry]):
    r"""Group is an entry that represent a group of other entries. For example,
    a "coreference group" is a group of coreferential entities. Each group will
    store a set of members, no duplications allowed.
    """
    MemberType: Type[Entry] = Entry

    def __init__(
        self,
        pack: PackType,
        members: Optional[Iterable[Entry]] = None,
    ):  # pylint: disable=useless-super-delegation
        self._members: Set[int] = set()
        super().__init__(pack, members)

    def add_member(self, member: Entry):
        r"""Add one entry to the group.

        Args:
            member: One member to be added to the group.
        """
        if not isinstance(member, self.MemberType):
            raise TypeError(
                f"The members of {type(self)} should be "
                f"instances of {self.MemberType}, but got {type(member)}"
            )
        self._members.add(member.tid)

    def get_members(self) -> List[Entry]:
        r"""Get the member entries in the group.

        Returns:
             A set of instances of :class:`~forte.data.ontology.core.Entry`
             that are the members of the group.
        """
        if self.pack is None:
            raise ValueError(
                "Cannot get members because group is not "
                "attached to any data pack."
            )
        member_entries = []
        for m in self._members:
            member_entries.append(self.pack.get_entry(m))
        return member_entries


class MultiPackGeneric(MultiEntry, Entry):
    def __init__(self, pack: PackType):
        super().__init__(pack=pack)


class MultiPackLink(MultiEntry, BaseLink):
    r"""This is used to link entries in a :class:`~forte.data.multi_pack.MultiPack`, which is
    designed to support cross pack linking, this can support applications such
    as sentence alignment and cross-document coreference. Each link should have
    a parent node and a child node. Note that the nodes are indexed by two
    integers, one additional index on which pack it comes from.
    """

    ParentType = Entry
    ChildType = Entry

    def __init__(
        self,
        pack: PackType,
        parent: Optional[Entry] = None,
        child: Optional[Entry] = None,
    ):
        self._parent: Optional[Tuple[int, int]] = None
        self._child: Optional[Tuple[int, int]] = None

        super().__init__(pack)

        if parent is not None:
            self.set_parent(parent)
        if child is not None:
            self.set_child(child)

    @property
    def parent(self) -> Tuple[int, int]:
        if self._parent is None:
            raise ValueError("Parent is not set for this link.")
        return self._parent

    @property
    def child(self) -> Tuple[int, int]:
        if self._child is None:
            raise ValueError("Child is not set for this link.")
        return self._child

    def parent_id(self) -> int:
        """
        Return the ``tid`` of the parent entry.

        Returns:
            The ``tid`` of the parent entry.
        """
        return self.parent[1]

    def child_id(self) -> int:
        """
        Return the ``tid`` of the child entry.

        Returns:
            The ``tid`` of the child entry.
        """
        return self.child[1]

    def parent_pack_id(self) -> int:
        """
        Return the `pack_id` of the parent pack.

        Returns:
            The `pack_id` of the parent pack..
        """
        if self._parent is None:
            raise ValueError("Parent is not set for this link.")
        return self.pack.packs[self._parent[0]].pack_id

    def child_pack_id(self) -> int:
        """
        Return the `pack_id` of the child pack.

        Returns:
            The `pack_id` of the child pack.
        """
        if self._child is None:
            raise ValueError("Child is not set for this link.")
        return self.pack.packs[self._child[0]].pack_id

    def set_parent(self, parent: Entry):
        r"""This will set the `parent` of the current instance with given Entry.
        The parent is saved internally as a tuple: ``pack index`` and
        ``entry.tid``. Pack index is the index of the data pack in the
        multi-pack.

        Args:
            parent: The parent of the link, which is an Entry from a data pack,
                it has access to the pack index and its own ``tid`` in the pack.
        """
        if not isinstance(parent, self.ParentType):
            raise TypeError(
                f"The parent of {type(self)} should be an "
                f"instance of {self.ParentType}, but get {type(parent)}"
            )
        # fix bug/enhancement #559: using pack_id instead of index
        # self._parent = self.pack.get_pack_index(parent.pack_id), parent.tid
        self._parent = parent.pack_id, parent.tid

    def set_child(self, child: Entry):
        r"""This will set the `child` of the current instance with given Entry.
        The child is saved internally as a tuple: ``pack index`` and
        ``entry.tid``. Pack index is the index of the data pack in the
        multi-pack.

        Args:
            child: The child of the link, which is an Entry from a data pack,
                it has access to the pack index and its own ``tid`` in the pack.
        """

        if not isinstance(child, self.ChildType):
            raise TypeError(
                f"The child of {type(self)} should be an "
                f"instance of {self.ChildType}, but get {type(child)}"
            )
        # fix bug/enhancement #559: using pack_id instead of index
        # self._child = self.pack.get_pack_index(child.pack_id), child.tid
        self._child = child.pack_id, child.tid

    def get_parent(self) -> Entry:
        r"""Get the parent entry of the link.

        Returns:
            An instance of :class:`~forte.data.ontology.core.Entry` that
            is the parent of the link.
        """
        if self._parent is None:
            raise ValueError("The parent of this link is not set.")

        pack_idx, parent_tid = self._parent
        return self.pack.get_subentry(pack_idx, parent_tid)

    def get_child(self) -> Entry:
        r"""Get the child entry of the link.

        Returns:
            An instance of :class:`~forte.data.ontology.core.Entry` that is
            the child of the link.
        """
        if self._child is None:
            raise ValueError("The parent of this link is not set.")

        pack_idx, child_tid = self._child
        return self.pack.get_subentry(pack_idx, child_tid)


# pylint: disable=duplicate-bases
class MultiPackGroup(MultiEntry, BaseGroup[Entry]):
    r"""Group type entries, such as "coreference group". Each group has a set
    of members.
    """
    MemberType: Type[Entry] = Entry

    def __init__(
        self, pack: PackType, members: Optional[Iterable[Entry]] = None
    ):  # pylint: disable=useless-super-delegation
        self._members: List[Tuple[int, int]] = []
        super().__init__(pack)
        if members is not None:
            self.add_members(members)

    def add_member(self, member: Entry):
        if not isinstance(member, self.MemberType):
            raise TypeError(
                f"The members of {type(self)} should be "
                f"instances of {self.MemberType}, but got {type(member)}"
            )

        self._members.append(
            # fix bug/enhancement 559: use pack_id instead of index
            (member.pack_id, member.tid)  # self.pack.get_pack_index(..)
        )

    def get_members(self) -> List[Entry]:
        members = []
        for pack_idx, member_tid in self._members:
            members.append(self.pack.get_subentry(pack_idx, member_tid))
        return members


@dataclass
class Query(Generics):
    r"""An entry type representing queries for information retrieval tasks.

    Args:
        pack: Data pack reference to which this query will be added
    """
    value: Optional[QueryType]
    results: Dict[str, float]

    def __init__(self, pack: PackType):
        super().__init__(pack)
        self.value: Optional[QueryType] = None
        self.results: Dict[str, float] = {}

    def add_result(self, pid: str, score: float):
        """
        Set the result score for a particular pack (based on the pack id).

        Args:
            pid: the pack id.
            score: the score for the pack

        Returns:
            None
        """
        self.results[pid] = score

    def update_results(self, pid_to_score: Dict[str, float]):
        r"""Updates the results for this query.

        Args:
             pid_to_score: A dict containing pack id -> score mapping
        """
        self.results.update(pid_to_score)


@total_ordering
class AudioAnnotation(Entry):
    r"""AudioAnnotation type entries, such as "recording" and "audio utterance".
    Each audio annotation has a :class:`~forte.data.span.Span` corresponding to its offset
    in the audio. Most methods in this class are the same as the ones in
    :class:`Annotation`, except that it replaces property `text` with `audio`.

    Args:
        pack: The container that this audio annotation
            will be added to.
        begin: The offset of the first sample in the audio annotation.
        end: The offset of the last sample in the audio annotation + 1.
    """

    def __init__(self, pack: PackType, begin: int, end: int):
        self._span: Optional[Span] = None
        self._begin: int = begin
        self._end: int = end
        super().__init__(pack)

    @property
    def audio(self):
        if self.pack is None:
            raise ValueError(
                "Cannot get audio because annotation is not "
                "attached to any data pack."
            )
        return self.pack.get_span_audio(self.begin, self.end)

    def __getstate__(self):
        r"""For serializing AudioAnnotation, we should create Span annotations
        for compatibility purposes.
        """
        self._span = Span(self._begin, self._end)
        state = super().__getstate__()
        state.pop("_begin")
        state.pop("_end")
        return state

    def __setstate__(self, state):
        """
        For de-serializing AudioAnnotation, we load the begin, end from Span,
        for compatibility purposes.
        """
        super().__setstate__(state)
        self._begin = self._span.begin
        self._end = self._span.end

    @property
    def span(self) -> Span:
        # Delay span creation at usage.
        if self._span is None:
            self._span = Span(self._begin, self._end)
        return self._span

    @property
    def begin(self):
        return self._begin

    @property
    def end(self):
        return self._end

    def __eq__(self, other):
        r"""The eq function of :class:`AudioAnnotation`.
        By default, :class:`AudioAnnotation` objects are regarded as the same if
        they have the same type, span, and are generated by the same component.

        Users can define their own eq function by themselves but this must
        be consistent to :meth:`hash`.
        """
        if other is None:
            return False
        return (type(self), self.begin, self.end) == (
            type(other),
            other.begin,
            other.end,
        )

    def __lt__(self, other):
        r"""To support total_ordering, `AudioAnnotation` must implement
        `__lt__`. The ordering is defined in the following way:

        1. If the begin of the audio annotations are different, the one with
           larger begin will be larger.
        2. In the case where the begins are the same, the one with larger
           end will be larger.
        3. In the case where both offsets are the same, we break the tie using
           the normal sorting of the class name.
        """
        if self.begin == other.begin:
            if self.end == other.end:
                return str(type(self)) < str(type(other))
            return self.end < other.end
        else:
            return self.begin < other.begin

    @property
    def index_key(self) -> int:
        return self.tid

    def get(
        self,
        entry_type: Union[str, Type[EntryType]],
        components: Optional[Union[str, Iterable[str]]] = None,
        include_sub_type=True,
    ) -> Iterable[EntryType]:
        """
        This function wraps the :meth:`~forte.data.data_pack.DataPack.get()` method to find
        entries "covered" by this audio annotation. See that method for more
        information. For usage details, refer to
        :meth:`forte.data.ontology.top.Annotation.get()`.

        Args:
            entry_type: The type of entries requested.
            components: The component (creator)
                generating the entries requested. If `None`, will return valid
                entries generated by any component.
            include_sub_type: whether to consider the sub types of
                the provided entry type. Default `True`.

        Yields:
            Each `Entry` found using this method.

        """
        yield from self.pack.get(entry_type, self, components, include_sub_type)


SinglePackEntries = (Link, Group, Annotation, Generics, AudioAnnotation)
MultiPackEntries = (MultiPackLink, MultiPackGroup, MultiPackGeneric)



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools as it
import logging
import os
from abc import ABC
from pathlib import Path
from typing import Optional, Any, List, Dict, Set, Tuple
from numpy import ndarray

from forte.data.ontology.code_generation_exceptions import (
    CodeGenerationException,
)
from forte.data.ontology.ontology_code_const import (
    SUPPORTED_PRIMITIVES,
    NON_COMPOSITES,
    COMPOSITES,
    Config,
    get_ignore_error_lines,
    AUTO_GEN_SIGNATURE,
    AUTO_GEN_FILENAME,
)
from forte.data.ontology.utils import split_file_path


class ImportManager:
    r"""A naive implementation that records import strings and imported names
    to be used. Mainly used to avoid import name conflicts such as:
       -- from user.module import token
       -- from system.module import token
    """

    def __init__(
        self, root: Optional["ImportManager"], module_name: Optional[str]
    ):
        self.__root = root
        self.__module_name = module_name
        self.__import_statements: List[str] = []
        # Defining names imported by this module.
        self.__imported_names: Dict[str, str] = {}
        # Defining names to be added by this generation module
        self.__defining_names: Dict[str, str] = {}
        self.__short_name_pool: Set[str] = set()
        self.__fix_modules = False

    def fix_modules(self):
        self.__fix_modules = True

    def is_known_name(self, full_class_name):
        """
        Check whether the class name can be used. It will check the class name
        in both the top manager or the current manager.

        Args:
            full_class_name: The name to be check.

        Returns: True if the class_name can be used, which means it is either
            imported or it is of a primitive type.

        """
        return (
            full_class_name in NON_COMPOSITES
            or full_class_name in COMPOSITES
            or self.is_imported(full_class_name)
            or full_class_name in self.__defining_names
        )

    def is_imported(self, class_name):
        """
        Check whether this `class_name` is already imported here in the module.
        Args:
            class_name: The name to be checked for importing.

        Returns: True if the class_name is imported.

        """
        if class_name in self.__imported_names:
            return True
        elif self.__root is None:
            return False
        else:
            if self.__root.is_imported(class_name) and not self.__fix_modules:
                self.add_object_to_import(class_name)
                return True

    def all_stored(self):
        return self.__imported_names.items()

    def get_name_to_use(self, full_name):
        if full_name in SUPPORTED_PRIMITIVES:
            return full_name

        if full_name in self.__defining_names:
            return self.__defining_names[full_name]

        if self.is_imported(full_name):
            return self.__imported_names[full_name]

    def get_import_statements(self):
        return sorted(self.__import_statements)

    def create_import_statement(self, full_name: str, as_name: str):
        if full_name not in NON_COMPOSITES:
            parts = full_name.split(".")
            class_name = parts[-1]

            if len(parts) > 1:
                module_name = ".".join(parts[:-1])

                if (
                    self.__module_name is None
                    or not module_name == self.__module_name
                ):
                    # No need to import classes in the same module
                    if class_name == as_name:
                        import_statement = (
                            f"from {module_name} " f"import {class_name}"
                        )
                    else:
                        import_statement = (
                            f"from {module_name} "
                            f"import {class_name} as {as_name}"
                        )

                    self.__import_statements.append(import_statement)
            else:
                if class_name == as_name:
                    import_statement = f"import {class_name}"
                else:
                    import_statement = f"import {class_name} as {as_name}"
                self.__import_statements.append(import_statement)

    def __find_next_available(self, class_name) -> str:
        counter = 0
        while True:
            as_name = f"{class_name}_{counter}"
            counter += 1
            if as_name not in self.__short_name_pool:
                break
        return as_name

    def __assign_as_name(self, full_name) -> str:
        class_name = full_name.split(".")[-1]
        if class_name not in self.__short_name_pool:
            self.__short_name_pool.add(class_name)
            return class_name
        else:
            as_name = self.__find_next_available(class_name)
            self.__short_name_pool.add(as_name)
            return as_name

    def add_defining_objects(self, full_name: str):
        if self.__fix_modules:
            # After fixing the modules, we should not add objects for import.
            raise CodeGenerationException(
                f"The module [{self.__module_name}] is fixed, cannot add "
                f"more objects."
            )

        if full_name not in self.__defining_names:
            class_name = full_name.split(".")[-1]
            if class_name not in self.__short_name_pool:
                self.__short_name_pool.add(class_name)
            else:
                logging.warning(
                    "Re-declared a new class named [%s]"
                    ", which is probably used in import.",
                    class_name,
                )
            self.__defining_names[full_name] = class_name

    def add_object_to_import(self, full_name: str):
        if self.__fix_modules:
            # After fixing the modules, we should not add objects for import.
            raise CodeGenerationException(
                f"The module [{self.__module_name}] is fixed, cannot add "
                f"more objects."
            )

        if full_name not in self.__imported_names:
            if full_name not in SUPPORTED_PRIMITIVES:
                as_name = self.__assign_as_name(full_name)
                self.__imported_names[full_name] = as_name
                self.create_import_statement(full_name, as_name)


class ImportManagerPool:
    """
    Holds import managers. The top_manager stores the default imports
    guaranteed by Forte's system ontology. The additional import managers are
    further populated by analyzing the generated code.
    """

    def __init__(self):
        self.__root_manager = ImportManager(None, None)
        self.__managers: Dict[str, ImportManager] = {}
        self.__default_imports: List[str] = []

    def add_default_import(self, full_name: str):
        self.__default_imports.append(full_name)

    @property
    def root(self) -> ImportManager:
        return self.__root_manager

    def get(self, module_name: str) -> ImportManager:
        if module_name in self.__managers:
            return self.__managers[module_name]
        else:
            nm = ImportManager(self.__root_manager, module_name)
            for full_name in self.__default_imports:
                nm.add_object_to_import(full_name)

            self.__managers[module_name] = nm
            return nm

    def fix_all_modules(self):
        self.__root_manager.fix_modules()
        for im in self.__managers.values():
            im.fix_modules()


def indent(level: int) -> str:
    return " " * Config.indent * level


def indent_line(line: str, level: int) -> str:
    return f"{indent(level)}{line}" if line else ""


def indent_code(
    code_lines: List[Optional[str]], level: int = 0, is_line_break: bool = True
) -> str:
    lines = []

    for code in code_lines:
        if code is None:
            continue
        if code == "":
            lines.append("")
        else:
            lines.extend(
                code.split(Config.line_break) if code is not None else []
            )
    ending = "\n" if is_line_break else ""
    return (
        Config.line_break.join([indent_line(line, level) for line in lines])
        + ending
    )


class EntryName:
    def __init__(self, entry_name: str):
        entry_splits = entry_name.split(".")
        self.filename, self.name = entry_splits[-2:]
        self.pkg = ".".join(entry_splits[0:-2])
        self.pkg_dir = self.pkg.replace(".", "/")
        self.module_name: str = f"{self.pkg}.{self.filename}"
        self.class_name: str = entry_name


class Item:
    def __init__(self, name: str, description: Optional[str]):
        self.name: str = name
        self.description: Optional[str] = description

    @property
    def field_name(self):
        return self.name

    def to_description(self, level: int) -> Optional[str]:
        if self.description is not None:
            return indent_code([self.description], level)
        # Returning a empty string will generate a placeholder for
        # the description.
        return ""


class Property(Item, ABC):
    def __init__(
        self,
        import_manager: ImportManager,
        name: str,
        type_str: str,
        description: Optional[str] = None,
        default_val: Any = None,
    ):
        super().__init__(name, description)
        self.type_str = type_str
        self.default_val = default_val
        self.import_manager: ImportManager = import_manager
        self.import_manager.add_object_to_import(self.type_str)

    def to_declaration(self, level: int):
        s = f"{self.field_name}: {self.internal_type_str()}"
        return indent_line(s, level)

    def internal_type_str(self) -> str:
        raise NotImplementedError

    def default_value(self) -> str:
        raise NotImplementedError

    def to_init_code(self, level: int) -> str:
        return indent_line(
            f"self.{self.field_name}: "
            f"{self.internal_type_str()} = "
            f"{self.default_value()}",
            level,
        )

    def to_description(self, level: int) -> Optional[str]:
        desc = f"{self.field_name} ({self.internal_type_str()}):"

        if self.description is not None and self.description.strip() != "":
            desc += f"\t{self.description}"
            return indent_line(desc, level)
        return indent_line(desc, level)

    def to_field_value(self):
        raise NotImplementedError


class ClassTypeDefinition:
    def __init__(
        self, name: str, type_str: str, description: Optional[str] = None
    ):
        self.name = name
        self.type_str = type_str
        self.description = description

    def to_code(self, level: int) -> str:
        return indent_code([f"{self.name} = {self.type_str}"], level, False)

    def to_field_value(self):
        pass


class NonCompositeProperty(Property):
    def __init__(
        self,
        import_manager: ImportManager,
        name: str,
        type_str: str,
        description: Optional[str] = None,
        default_val: Any = None,
        self_ref: bool = False,
    ):
        super().__init__(
            import_manager, name, type_str, description, default_val
        )

        # Primitive type will use optional in type string, so we add the
        # optional here.
        self.option_type = "typing.Optional"
        import_manager.add_object_to_import(self.option_type)

        self.is_forte_type = import_manager.is_imported(type_str)
        self.self_ref = self_ref

    def internal_type_str(self) -> str:
        option_type = self.import_manager.get_name_to_use(self.option_type)
        type_str = self.import_manager.get_name_to_use(self.type_str)
        if self.self_ref:
            type_str = f"'{type_str}'"
        return f"{option_type}[{type_str}]"

    def default_value(self) -> str:
        return repr(self.default_val)

    def to_field_value(self):
        return self.name


class NdArrayProperty(Property):
    """
    NdArrayProperty accepts parsed properties of NdArray and
    instructs import manager to import and instanciate FNdArray
    as default value in the generated code.
    """

    def __init__(
        self,
        import_manager: ImportManager,
        name: str,
        ndarray_dtype: Optional[str] = None,
        ndarray_shape: Optional[List[int]] = None,
        description: Optional[str] = None,
        default_val: Optional[ndarray] = None,
    ):
        self.type_str = "forte.data.ontology.core.FNdArray"
        super().__init__(
            import_manager,
            name,
            self.type_str,
            description=description,
            default_val=default_val,
        )
        self.ndarray_dtype: Optional[str] = ndarray_dtype
        self.ndarray_shape: Optional[List[int]] = ndarray_shape

    def internal_type_str(self) -> str:
        type_str = self.import_manager.get_name_to_use(self.type_str)
        return f"{type_str}"

    def default_value(self) -> str:
        if self.ndarray_dtype is None:
            return f"FNdArray(shape={self.ndarray_shape}, dtype={self.ndarray_dtype})"
        else:
            return f"FNdArray(shape={self.ndarray_shape}, dtype='{self.ndarray_dtype}')"

    def _full_class(self):
        item_type = self.import_manager.get_name_to_use(self.type_str)
        return item_type

    def to_field_value(self):
        return self.name


class DictProperty(Property):
    def __init__(
        self,
        import_manager: ImportManager,
        name: str,
        key_type: str,
        value_type: str,
        description: Optional[str] = None,
        default_val: Any = None,
        self_ref: bool = False,
    ):
        if not key_type == "str":
            # This string value constraint is to conform with JSON format
            #  requirement: https://www.json.org/json-en.html
            raise CodeGenerationException(
                f"Dictionary keys can only be string values, find {key_type} "
                f"at {name}."
            )

        self.value_is_forte_type = import_manager.is_imported(value_type)
        type_str = (
            "forte.data.ontology.core.FDict"
            if self.value_is_forte_type
            else "typing.Dict"
        )
        super().__init__(
            import_manager,
            name,
            type_str,
            description=description,
            default_val=default_val,
        )
        self.key_type: str = key_type
        self.value_type: str = value_type
        self.self_ref: bool = self_ref

    def internal_type_str(self) -> str:
        # option_type = self.import_manager.get_name_to_use('typing.Optional')
        return f"{self._full_class()}"

    def default_value(self) -> str:
        if self.type_str == "typing.Dict":
            return "dict()"
        else:
            return "FDict(self)"

    def _full_class(self) -> str:
        composite_type = self.import_manager.get_name_to_use(self.type_str)
        key_type = self.import_manager.get_name_to_use(self.key_type)
        value_type = self.import_manager.get_name_to_use(self.value_type)
        if self.self_ref:
            value_type = f"'{value_type}'"

        return f"{composite_type}[{key_type}, {value_type}]"

    def to_field_value(self):
        return self.name


class ListProperty(Property):
    def __init__(
        self,
        import_manager: ImportManager,
        name: str,
        item_type: str,
        description: Optional[str] = None,
        default_val: Any = None,
        self_ref: bool = False,
    ):
        self.value_is_forte_type = import_manager.is_imported(item_type)
        type_str = (
            "forte.data.ontology.core.FList"
            if self.value_is_forte_type
            else "typing.List"
        )
        super().__init__(
            import_manager,
            name,
            type_str,
            description=description,
            default_val=default_val,
        )
        self.item_type: str = item_type
        self.self_ref: bool = self_ref

    def internal_type_str(self) -> str:
        # option_type = self.import_manager.get_name_to_use('typing.Optional')
        return f"{self._full_class()}"

    def default_value(self) -> str:
        if self.type_str == "typing.List":
            return "[]"
        else:
            return "FList(self)"

    def _full_class(self):
        composite_type = self.import_manager.get_name_to_use(self.type_str)
        item_type = self.import_manager.get_name_to_use(self.item_type)
        if self.self_ref:
            item_type = f"'{item_type}'"

        return f"{composite_type}[{item_type}]"

    def to_field_value(self):
        return self.name


class EntryDefinition(Item):
    def __init__(
        self,
        name: str,
        class_type: str,
        init_args: Optional[str] = None,
        properties: Optional[List[Property]] = None,
        class_attributes: Optional[List[ClassTypeDefinition]] = None,
        description: Optional[str] = None,
    ):
        super().__init__(name, description)
        self.class_type = class_type
        self.properties: List[Property] = (
            [] if properties is None else properties
        )
        self.class_attributes = (
            [] if class_attributes is None else class_attributes
        )
        self.description = description if description else None
        self.init_args = init_args if init_args is not None else ""
        self.init_args = self.init_args.replace("=", " = ")

    def to_init_code(self, level: int) -> str:
        return indent_line(f"def __init__(self, {self.init_args}):", level)

    def to_property_code(self, level: int) -> str:
        lines = []
        for p in self.properties:
            lines.append(p.to_declaration(0))
        return indent_code(
            [indent_line(line, 0) for line in lines], level, False
        )

    def to_class_attribute_code(self, level: int):
        lines = [item.to_code(0) for item in self.class_attributes]
        return indent_code(
            [indent_line(line, 0) for line in lines], level, False
        )

    def to_code(self, level: int) -> str:
        super_args = ", ".join(
            [item.split(":")[0].strip() for item in self.init_args.split(",")]
        )
        raw_desc = self.to_description(1)
        desc: str = "" if raw_desc is None else raw_desc
        lines: List[Optional[str]] = [
            "",
            "",
            "@dataclass",
            f"class {self.name}({self.class_type}):",
        ]

        lines += [desc] if desc.strip() else []

        lines.append("")

        property_code = self.to_property_code(1)
        if property_code:
            lines.append(property_code)
            lines.append("")

        class_attr_code = self.to_class_attribute_code(1)
        if class_attr_code:
            lines.append(class_attr_code)
            lines.append("")

        lines += [
            self.to_init_code(1),
            indent_line(f"super().__init__({super_args})", 2),
        ]
        lines += [item.to_init_code(2) for item in self.properties]
        lines += [""]

        return indent_code(lines, level, False)

    @staticmethod
    def to_item_descs(items, title):
        item_descs = [item.to_description(0) for item in items]
        item_descs = [item for item in item_descs if item is not None]
        if len(item_descs) > 0:
            item_descs = [indent_line(title, 0)] + [
                indent_line(desc, 1) for desc in item_descs
            ]
        return item_descs

    def to_description(self, level: int) -> Optional[str]:
        class_desc = [] if self.description is None else [self.description + ""]
        item_descs = self.to_item_descs(self.properties, "Attributes:")

        descs = class_desc + item_descs
        if len(descs) == 0:
            return ""
        quotes = indent_line('"""', 0)
        return indent_code([quotes] + descs + [quotes], level, False)


class ModuleWriter:
    """
    A writer to write entry definitions to a file.
    """

    def __init__(self, module_name: str, import_managers: ImportManagerPool):
        self.module_name = module_name
        self.source_file: str = ""

        self.description: Optional[str] = None
        self.import_managers: ImportManagerPool = import_managers
        self.entries: List[Tuple[EntryName, EntryDefinition]] = []

        pkg, self.file_name = self.module_name.rsplit(".", 1)
        self.pkg_dir = pkg.replace(".", "/")

    def set_description(self, description: str):
        self.description = description

    def add_entry(self, entry_name: EntryName, entry_item: EntryDefinition):
        self.entries.append((entry_name, entry_item))

    def make_module_dirs(
        self,
        tempdir: str,
        destination: str,
        namespace_depth: int,
    ):
        """
        Create entry sub-directories with .generated file to indicate the
         subdirectory is created by this procedure. No such file will be added
         if the directory already exists.

        Args:
            tempdir: A temp directory to create the structure, code will be
              first generated here.
            destination: The destination directory where the code should be
              placed
            namespace_depth: set an integer argument namespace_depth to allow
              customized number of levels of namespace packaging.
              The generation of __init__.py for all the directory
              levels above namespace_depth will be disabled.
              For example, if we have an ontology level1.levle2.level3.
              something and namespace_depth=2, then we remove __init__.py
              under level1 and level1/level2 while keeping __init__.py under
              level1/level2/level3.
              When namespace_depth<=0, we just disable namespace packaging
              and include __init__.py in all directory levels.
        Returns:
        """
        entry_dir_split = split_file_path(self.pkg_dir)

        rel_dir_paths = it.accumulate(entry_dir_split, os.path.join)
        count = 1
        for rel_dir_path in rel_dir_paths:
            temp_path = os.path.join(tempdir, rel_dir_path)
            if not os.path.exists(temp_path):
                os.mkdir(temp_path)

            dest_path = os.path.join(destination, rel_dir_path)
            dest_path_exists = os.path.exists(dest_path)
            if not dest_path_exists:
                Path(os.path.join(temp_path, AUTO_GEN_FILENAME)).touch()

            # Create init file
            if count > namespace_depth:
                if not dest_path_exists:
                    init_file_path = os.path.join(temp_path, "__init__.py")
                    with open(
                        init_file_path, "w", encoding="utf-8"
                    ) as init_file:
                        init_file.write(f"# {AUTO_GEN_SIGNATURE}\n")
            count += 1

    def write(
        self,
        tempdir: str,
        destination: str,
        namespace_depth: int,
    ):
        """
        Write the entry information to file.

        Args:
            tempdir: A temporary directory for writing intermediate files.
            destination: The actual folder to place the generated code.
            namespace_depth: set an integer argument namespace_depth to allow
              customized number of levels of namespace packaging.
              The generation of __init__.py for all the directory
              levels above namespace_depth will be disabled.
              For example, if we have an ontology level1.levle2.level3.
              something and namespace_depth=2, then we remove __init__.py
              under level1 and level1/level2 while keeping __init__.py under
              level1/level2/level3.
              When namespace_depth<=0, we just disable namespace packaging
              and include __init__.py in all directory levels.
        Returns:

        """

        self.make_module_dirs(tempdir, destination, namespace_depth)
        full_path = os.path.join(tempdir, self.pkg_dir, self.file_name) + ".py"

        with open(full_path, "w", encoding="utf-8") as f:
            # Write header.
            f.write(self.to_header(0))
            for entry_name, entry_item in self.entries:
                logging.info("Writing class: %s", entry_name.class_name)
                f.write(entry_item.to_code(0))

    def to_header(self, level: int) -> str:
        all_first_line = indent_line("__all__ = [", 0)
        all_mid_lines = indent_code(
            [f'"{en.name}",' for en, ei in self.entries], 1, False
        )
        all_last_line = indent_line("]", 0)

        lines = [
            self.to_description(0),
            self.to_import_code(0),
            all_first_line,
            all_mid_lines,
            all_last_line,
        ]
        return indent_code(lines, level)

    def to_description(self, level):
        quotes = '"""'
        lines = get_ignore_error_lines(self.source_file) + [
            quotes,
            self.description,
            quotes,
        ]
        return indent_code(lines, level)

    def to_import_code(self, level):
        return indent_code(
            self.import_managers.get(self.module_name).get_import_statements(),
            level,
        )


class ModuleWriterPool:
    def __init__(self, import_managers: ImportManagerPool):
        self.__module_writers: Dict[str, ModuleWriter] = {}
        self.__import_managers = import_managers

    def get(self, module_name: str) -> ModuleWriter:
        if module_name in self.__module_writers:
            return self.__module_writers[module_name]
        else:
            mw = ModuleWriter(module_name, self.__import_managers)
            self.__module_writers[module_name] = mw
            return mw

    def writers(self):
        return self.__module_writers.values()


class EntryTreeNode:
    def __init__(self, name: str):
        self.children: List[EntryTreeNode] = []
        self.parent: Optional[EntryTreeNode] = None
        self.name: str = name
        self.attributes: Set[str] = set()

    def __repr__(self):
        r"""for printing purpose."""
        attr_str = ", ".join(self.attributes)
        return self.name + ": " + attr_str


class EntryTree:
    r"""
    A tree structure based on the parent-children relations of the entries.
    """

    def __init__(self):
        self.root = EntryTreeNode("root")

    def add_node(
        self,
        curr_entry_name: str,
        parent_entry_name: str,
        curr_entry_attr: Set[str],
    ):
        r"""Add a tree node with `curr_entry_name` as a child to
        `parent_entry_name` in the tree, the attributes `curr_entry_attr`
        would be added to the tree node attributes.

        Args:
            curr_entry_name: the type name of the node to be added.
            parent_entry_name: the type name of the parent of the node to be
                added.
            curr_entry_attr: the attributes of the node to be added.

        Returns:

        """
        found_node = search(self.root, curr_entry_name)
        if found_node is None:
            curr_entry_node = EntryTreeNode(curr_entry_name)
            curr_entry_node.attributes = curr_entry_attr
            parent_in_tree = search(self.root, parent_entry_name)
            if parent_in_tree is None:
                parent_in_tree = EntryTreeNode(parent_entry_name)
                self.root.children.append(parent_in_tree)
                parent_in_tree.parent = self.root
            parent_in_tree.children.append(curr_entry_node)
            curr_entry_node.parent = parent_in_tree
        else:
            found_node.attributes = curr_entry_attr

    def print_traverse(self):
        path = []
        traverse(self.root, path)

    def collect_parents(self, node_dict: Dict[str, Set[str]]):
        r"""Collect all the parent nodes for all the nodes in the `node_dict`
        and add the types and attributes of these parent nodes to `node_dict`.

        Args:
            node_dict: the nodes dictionary of nodes to collect parent nodes
                for.

        """
        input_node_dict = node_dict.copy()
        for node_name in input_node_dict.keys():
            found_node = search(self.root, search_node_name=node_name)
            if found_node is not None:
                while found_node.parent.name != "root":
                    node_dict[
                        found_node.parent.name
                    ] = found_node.parent.attributes
                    found_node = found_node.parent

    def todict(self) -> Dict[str, Any]:
        r"""Dump the EntryTree structure to a dictionary.

        Returns:
            dict: A dictionary storing the EntryTree.
        """

        def node_to_dict(node: EntryTreeNode):
            return (
                None
                if not node
                else {
                    "name": node.name,
                    "attributes": list(node.attributes),
                    "children": [
                        node_to_dict(child) for child in node.children
                    ],
                }
            )

        return node_to_dict(self.root)

    def fromdict(
        self, tree_dict: Dict[str, Any], parent_entry_name: Optional[str] = None
    ) -> Optional["EntryTree"]:
        r"""Load the EntryTree structure from a dictionary.

        Args:
            tree_dict: A dictionary storing the EntryTree.
            parent_entry_name: The type name of the parent of the node to be
                built. Default value is None.
        """
        if not tree_dict:
            return None

        if parent_entry_name is None:
            self.root = EntryTreeNode(name=tree_dict["name"])
            self.root.attributes = set(tree_dict["attributes"])
        else:
            self.add_node(
                curr_entry_name=tree_dict["name"],
                parent_entry_name=parent_entry_name,
                curr_entry_attr=set(tree_dict["attributes"]),
            )
        for child in tree_dict["children"]:
            self.fromdict(child, tree_dict["name"])
        return self


def search(node: EntryTreeNode, search_node_name: str):
    if node.name == search_node_name:
        return node

    for child in node.children:
        tmp = search(child, search_node_name)
        if tmp:
            return tmp


def traverse(node: EntryTreeNode, path: List[str]):
    path.append(repr(node))
    if len(node.children) == 0:
        print(path)
        path.pop()
    else:
        for child in node.children:
            traverse(child, path)
        path.pop()



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=logging-fstring-interpolation
import logging
import os
from typing import Dict, List, Optional, Tuple
import numpy as np
from forte.utils import create_import_error_msg
from forte.common.configuration import Config
from forte.common.resources import Resources
from forte.data.data_pack import DataPack
from forte.data.ontology import Annotation
from forte.models.ner import utils
from forte.models.ner.model_factory import BiRecurrentConvCRF
from forte.processors.base.batch_processor import RequestPackingProcessor
from ft.onto.base_ontology import Token, EntityMention

try:
    import torch
except ImportError as e:
    raise ImportError(
        create_import_error_msg("torch", "models", "ner predictor")
    ) from e


logger = logging.getLogger(__name__)

__all__ = [
    "CoNLLNERPredictor",
]


class CoNLLNERPredictor(RequestPackingProcessor):
    """
    An Named Entity Recognizer trained according to `Ma, Xuezhe, and Eduard
    Hovy. "End-to-end sequence labeling via bi-directional lstm-cnns-crf."
    <https://arxiv.org/abs/1603.01354>`_.

    Note that to use :class:`CoNLLNERPredictor`, the :attr:`ontology` of
    :class:`~forte.pipeline.Pipeline` must be an ontology that include
    :class:`ft.onto.base_ontology.Token` and :class:`ft.onto.base_ontology.Sentence`.
    """

    def __init__(self):
        super().__init__()
        self.model = None
        self.word_alphabet, self.char_alphabet, self.ner_alphabet = (
            None,
            None,
            None,
        )
        self.resource = None
        self.config_model = None
        self.config_data = None
        self.normalize_func = None
        self.device = None

    def initialize(self, resources: Resources, configs: Config):
        super().initialize(resources, configs)

        self.resource = resources
        self.config_model = configs.config_model
        self.config_data = configs.config_data

        resource_path = configs.config_model.resource_dir

        keys = {
            "word_alphabet",
            "char_alphabet",
            "ner_alphabet",
            "word_embedding_table",
        }

        missing_keys = list(keys.difference(self.resource.keys()))

        self.resource.load(keys=missing_keys, path=resource_path)

        self.word_alphabet = resources.get("word_alphabet")
        self.char_alphabet = resources.get("char_alphabet")
        self.ner_alphabet = resources.get("ner_alphabet")
        word_embedding_table = resources.get("word_embedding_table")

        if resources.get("device"):
            self.device = resources.get("device")
        else:
            self.device = (
                torch.device("cuda")
                if torch.cuda.is_available()
                else torch.device("cpu")
            )

        self.normalize_func = utils.normalize_digit_word

        if "model" not in self.resource.keys():

            def load_model(path):
                model = BiRecurrentConvCRF(
                    word_embedding_table,
                    self.char_alphabet.size(),
                    self.ner_alphabet.size(),
                    self.config_model,
                )

                if os.path.exists(path):
                    with open(path, "rb") as f:
                        weights = torch.load(f, map_location=self.device)
                        model.load_state_dict(weights)
                return model

            self.resource.load(keys={"model": load_model}, path=resource_path)

        self.model = resources.get("model")
        self.model.to(self.device)
        self.model.eval()

        utils.set_random_seed(self.config_model.random_seed)

    @torch.no_grad()
    def predict(
        self, data_batch: Dict[str, Dict[str, List[str]]]
    ) -> Dict[str, Dict[str, List[np.ndarray]]]:
        tokens = data_batch["Token"]

        instances = []
        for words in tokens["text"]:
            char_id_seqs = []
            word_ids = []
            for word in words:
                char_ids = []
                for char in word:
                    char_ids.append(self.char_alphabet.get_index(char))
                if len(char_ids) > self.config_data.max_char_length:
                    char_ids = char_ids[: self.config_data.max_char_length]
                char_id_seqs.append(char_ids)

                word = self.normalize_func(word)
                word_ids.append(self.word_alphabet.get_index(word))

            instances.append((word_ids, char_id_seqs))

        self.model.eval()
        batch_data = self.get_batch_tensor(instances, device=self.device)
        word, char, masks, unused_lengths = batch_data
        preds = self.model.decode(word, char, mask=masks)

        pred: Dict = {"Token": {"ner": [], "tid": []}}

        for i in range(len(tokens["tid"])):
            tids = tokens["tid"][i]
            ner_tags = []
            for j in range(len(tids)):
                ner_tags.append(self.ner_alphabet.get_instance(preds[i][j]))

            pred["Token"]["ner"].append(np.array(ner_tags))
            pred["Token"]["tid"].append(np.array(tids))

        return pred

    def load_model_checkpoint(self, model_path=None):
        p = (
            model_path
            if model_path is not None
            else self.config_model.model_path
        )
        ckpt = torch.load(p, map_location=self.device)
        logger.info(f"Restoring NER model from {self.config_model.model_path}")
        self.model.load_state_dict(ckpt["model"])

    def pack(
        self,
        pack: DataPack,
        predict_results: Dict[str, Dict[str, List[str]]],
        _: Optional[Annotation] = None,
    ):
        """
        Write the prediction results back to datapack. by writing the predicted
        ner to the original tokens.
        """

        if predict_results is None:
            return

        current_entity_mention: Tuple[int, str] = (-1, "None")

        for i in range(len(predict_results["Token"]["tid"])):
            # an instance
            for j in range(len(predict_results["Token"]["tid"][i])):
                tid: int = predict_results["Token"]["tid"][i][j]  # type: ignore

                orig_token: Token = pack.get_entry(tid)  # type: ignore
                ner_tag: str = predict_results["Token"]["ner"][i][j]

                orig_token.ner = ner_tag

                token = orig_token
                token_ner = token.ner
                assert isinstance(token_ner, str)
                if token_ner[0] == "B":
                    current_entity_mention = (token.begin, token_ner[2:])
                elif token_ner[0] == "I":
                    continue
                elif token_ner[0] == "O":
                    continue

                elif token_ner[0] == "E":
                    if token_ner[2:] != current_entity_mention[1]:
                        continue

                    entity = EntityMention(
                        pack, current_entity_mention[0], token.end
                    )
                    entity.ner_type = current_entity_mention[1]
                elif token_ner[0] == "S":
                    current_entity_mention = (token.begin, token_ner[2:])
                    entity = EntityMention(
                        pack, current_entity_mention[0], token.end
                    )
                    entity.ner_type = current_entity_mention[1]

    def get_batch_tensor(
        self,
        data: List[Tuple[List[int], List[List[int]]]],
        device: Optional[torch.device] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """Get the tensors to be fed into the model.

        Args:
            data: A list of tuple (word_ids, char_id_sequences)
            device: The device for the tensors.

        Returns:
            A tuple where

            - ``words``: A tensor of shape `[batch_size, batch_length]`
              representing the word ids in the batch
            - ``chars``: A tensor of shape
              `[batch_size, batch_length, char_length]` representing the char
              ids for each word in the batch
            - ``masks``: A tensor of shape `[batch_size, batch_length]`
              representing the indices to be masked in the batch. 1 indicates
              no masking.
            - ``lengths``: A tensor of shape `[batch_size]` representing the
              length of each sentences in the batch
        """
        batch_size = len(data)
        batch_length = max([len(d[0]) for d in data])
        char_length = max(
            [max([len(charseq) for charseq in d[1]]) for d in data]
        )

        char_length = min(
            self.config_data.max_char_length,
            char_length + self.config_data.num_char_pad,
        )

        wid_inputs = np.empty([batch_size, batch_length], dtype=np.int64)
        cid_inputs = np.empty(
            [batch_size, batch_length, char_length], dtype=np.int64
        )

        masks = np.zeros([batch_size, batch_length], dtype=np.float32)

        lengths = np.empty(batch_size, dtype=np.int64)

        for i, inst in enumerate(data):
            wids, cid_seqs = inst

            inst_size = len(wids)
            lengths[i] = inst_size
            # word ids
            wid_inputs[i, :inst_size] = wids
            wid_inputs[i, inst_size:] = self.word_alphabet.pad_id
            for c, cids in enumerate(cid_seqs):
                cid_inputs[i, c, : len(cids)] = cids
                cid_inputs[i, c, len(cids) :] = self.char_alphabet.pad_id
            cid_inputs[i, inst_size:, :] = self.char_alphabet.pad_id
            masks[i, :inst_size] = 1.0

        words = torch.from_numpy(wid_inputs).to(device)
        chars = torch.from_numpy(cid_inputs).to(device)
        masks = torch.from_numpy(masks).to(device)
        lengths = torch.from_numpy(lengths).to(device)

        return words, chars, masks, lengths

    # TODO: change this to manageable size
    @classmethod
    def default_configs(cls):
        r"""Default config for NER Predictor"""
        return {
            "config_data": {
                "train_path": "",
                "val_path": "",
                "test_path": "",
                "num_epochs": 200,
                "batch_size_tokens": 512,
                "test_batch_size": 16,
                "max_char_length": 45,
                "num_char_pad": 2,
            },
            "config_model": {
                "output_hidden_size": 128,
                "dropout_rate": 0.3,
                "word_emb": {"dim": 100},
                "char_emb": {"dim": 30, "initializer": {"type": "normal_"}},
                "char_cnn_conv": {
                    "in_channels": 30,
                    "out_channels": 30,
                    "kernel_size": 3,
                    "padding": 2,
                },
                "bilstm_sentence_encoder": {
                    "rnn_cell_fw": {
                        "input_size": 130,
                        "type": "LSTMCell",
                        "kwargs": {"num_units": 128},
                    },
                    "rnn_cell_share_config": "yes",
                    "output_layer_fw": {"num_layers": 0},
                    "output_layer_share_config": "yes",
                },
                "learning_rate": 0.01,
                "momentum": 0.9,
                "decay_interval": 1,
                "decay_rate": 0.05,
                "random_seed": 1234,
                "initializer": {"type": "xavier_uniform_"},
                "model_path": "",
                "resource_dir": "",
            },
            "batcher": {
                "batch_size": 16,
                "context_type": "ft.onto.base_ontology.Sentence",
                "requests": {
                    "ft.onto.base_ontology.Token": [],
                    "ft.onto.base_ontology.Sentence": [],
                },
            },
        }



# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Class for data augmentation algorithm. The text replacement op will
replace a piece of text with data augmentation algorithms.
"""
from typing import Tuple, Union, Dict, Any
from abc import abstractmethod, ABC
from forte.data.ontology.top import Annotation
from forte.common.configuration import Config

__all__ = [
    "TextReplacementOp",
]


class TextReplacementOp(ABC):
    r"""
    The base class holds the data augmentation algorithm.
    We leave the :func:`replace` method to be implemented
    by subclasses.
    """

    def __init__(self, configs: Union[Config, Dict[str, Any]]):
        r"""
        Set the configuration for the text replacement op.
        """
        self.configs: Config = Config(configs, None, allow_new_hparam=True)

    @abstractmethod
    def replace(self, input_anno: Annotation) -> Tuple[bool, str]:
        r"""
        Most data augmentation algorithms can be considered
        as replacement-based methods on different levels.
        This function takes in an annotation as input and
        returns the augmented string.

        Args:
            input_anno: the input annotation to be replaced.

        Returns:
            A tuple, where the first element is a boolean value indicating
            whether the replacement happens, and the second element is the
            replaced string.
        """
        raise NotImplementedError



# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The reader that reads AG News data into Datapacks.
"""
from typing import Iterator, Tuple

from ftx.onto.ag_news import Description
from ft.onto.base_ontology import Document, Title
from forte.data.data_pack import DataPack
from forte.data.base_reader import PackReader


__all__ = [
    "AGNewsReader",
]


class AGNewsReader(PackReader):
    r""":class:`AGNewsReader` is designed to read in AG News
    text classification dataset.
    The AG's news topic classification dataset is constructed by Xiang Zhang
    (xiang.zhang@nyu.edu) from the AG corpus. It is used as a text
    classification benchmark in the following paper:
    https://arxiv.org/abs/1509.01626
    The dataset can be downloaded from:
    https://github.com/mhjabreel/CharCnn_Keras/tree/master/data/ag_news_csv

    The input to this reader is the path to the CSV file.
    """

    def _collect(  # type: ignore
        self, csv_file: str
    ) -> Iterator[Tuple[int, str]]:
        r"""Collects from a CSV file path and returns an iterator of AG News
        data. The elements in the iterator correspond to each line
        in the csv file. One line is expected to be parsed as one
        DataPack.

        Args:
            csv_file: A CSV file path.

        Returns: Iterator of each line in the csv file.
        """
        with open(csv_file, "r", encoding="utf-8") as f:
            for line_id, line in enumerate(f):
                yield line_id, line

    def _cache_key_function(self, line_info: Tuple[int, str]) -> str:
        return str(line_info[0])

    def _parse_pack(self, line_info: Tuple[int, str]) -> Iterator[DataPack]:
        line_id, line = line_info

        pack = DataPack()
        text: str = ""
        line = line.strip()
        data = line.split(",")

        class_id: int = int(data[0].replace('"', ""))
        title: str = data[1]
        description: str = data[2]

        text += title
        title_end = len(text)
        text += "\n" + description
        description_start = title_end + 1
        description_end = len(text)

        pack.set_text(text, replace_func=self.text_replace_operation)

        doc = Document(pack, 0, description_end)
        doc.document_class = [self.configs.class_names[class_id]]
        Title(pack, 0, title_end)
        Description(pack, description_start, description_end)

        pack.pack_name = line_id
        yield pack

    @classmethod
    def default_configs(cls):
        config: dict = super().default_configs()

        config.update(
            {
                "class_names": {
                    1: "World",
                    2: "Sports",
                    3: "Business",
                    4: "Sci/Tech",
                }
            }
        )
        return config



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The reader that reads CoNLL-U Standard Universal Dependency Format -
https://universaldependencies.org/docs/format.html
into data_pack format
"""
from typing import Iterator, Dict, Tuple, Any

from ft.onto.base_ontology import (
    Document,
    Sentence,
    Token,
    Dependency,
    EnhancedDependency,
)

from forte.data.data_utils_io import dataset_path_iterator
from forte.data.data_pack import DataPack
from forte.data.base_reader import PackReader

__all__ = ["ConllUDReader"]


class ConllUDReader(PackReader):
    r""":class:`~forte.data.readers.conllu_ud_reader.ConllUDReader` is designed
    to read in the Universal Dependencies
    2.4 dataset.
    """

    def _cache_key_function(self, data_pack: Any) -> str:
        if data_pack.pack_name is None:
            raise ValueError("data_pack does not have a document id")
        return data_pack.pack_name

    def _collect(self, *args, **kwargs) -> Iterator[Any]:
        # pylint: disable = unused-argument
        r"""Iterator over conll files in the data_source.

        Args:
            args: args[0] is the directory to the conllu files.
            kwargs:

        Returns: data packs obtained from each document from each conllu file.
        """
        conll_dir_path = args[0]
        file_paths = dataset_path_iterator(conll_dir_path, "conllu")
        for file_path in file_paths:
            with open(file_path, "r", encoding="utf8") as file:
                lines = file.readlines()
                doc_lines = []

                for i, line in enumerate(lines):
                    # previous document ends
                    doc_lines.append(line)
                    if i == len(lines) - 1 or lines[i + 1].strip().startswith(
                        "# newdoc"
                    ):
                        yield doc_lines
                        doc_lines = []

    def _parse_pack(self, doc_lines) -> Iterator[DataPack]:
        token_comp_fields = [
            "id",
            "form",
            "lemma",
            "pos",
            "ud_xpos",
            "ud_features",
            "head",
            "label",
            "enhanced_dependency_relations",
            "ud_misc",
        ]

        token_multi_fields = [
            "ud_features",
            "ud_misc",
            "enhanced_dependency_relations",
        ]

        token_feature_fields = ["ud_features", "ud_misc"]

        data_pack: DataPack = DataPack()
        doc_sent_begin: int = 0
        doc_num_sent: int = 0
        doc_text: str = ""
        doc_offset: int = 0
        doc_id: str

        sent_text: str
        sent_tokens: Dict[str, Tuple[Dict[str, Any], Token]] = {}

        for line in doc_lines:
            line = line.strip()
            line_comps = line.split()

            if line.startswith("# newdoc"):
                doc_id = line.split("=")[1].strip()

            elif line.startswith("# sent"):
                sent_text = ""

            elif len(line_comps) > 0 and line_comps[0].strip().isdigit():
                # token
                token_comps: Dict[str, Any] = {}

                for index, key in enumerate(token_comp_fields):
                    token_comps[key] = str(line_comps[index])

                    if key in token_multi_fields:
                        values = (
                            str(token_comps[key]).split("|")
                            if token_comps[key] != "_"
                            else []
                        )
                        if key not in token_feature_fields:
                            token_comps[key] = values
                        else:
                            feature_lst = [
                                elem.split("=", 1) for elem in values
                            ]
                            feature_dict = {
                                elem[0]: elem[1] for elem in feature_lst
                            }
                            token_comps[key] = feature_dict

                word: str = token_comps["form"]
                word_begin = doc_offset
                word_end = doc_offset + len(word)

                # add token
                token: Token = Token(data_pack, word_begin, word_end)

                token.lemma = token_comps["lemma"]
                token.pos = token_comps["pos"]
                token.ud_xpos = token_comps["ud_xpos"]
                token.ud_features = token_comps["ud_features"]
                token.ud_misc = token_comps["ud_misc"]

                sent_tokens[str(token_comps["id"])] = (token_comps, token)

                sent_text += word + " "
                doc_offset = word_end + 1

            elif line == "":
                # sentence ends
                sent_text = sent_text.strip()
                doc_text += " " + sent_text

                # add dependencies for a sentence when all the tokens have been
                # added
                for _, (token_comps, token) in sent_tokens.items():
                    # add primary dependency
                    label = token_comps["label"]
                    if label == "root":
                        token.is_root = True
                    else:
                        token.is_root = False
                        head = sent_tokens[token_comps["head"]][1]
                        dependency = Dependency(data_pack, head, token)
                        dependency.dep_label = label

                    # add enhanced dependencies
                    for dep in token_comps["enhanced_dependency_relations"]:
                        head_id, label = dep.split(":", 1)
                        if label != "root":
                            head = sent_tokens[head_id][1]
                            enhanced_dependency = EnhancedDependency(
                                data_pack, head, token
                            )
                            enhanced_dependency.dep_label = label

                # add sentence
                Sentence(data_pack, doc_sent_begin, doc_offset - 1)

                doc_sent_begin = doc_offset
                doc_num_sent += 1

        doc_text = doc_text.strip()
        data_pack.set_text(doc_text)

        # add doc to data_pack
        Document(data_pack, 0, len(doc_text))
        data_pack.pack_name = doc_id

        yield data_pack



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
__all__ = ["LabeledSpanGraphNetwork"]
import math
from collections import defaultdict
from typing import Any, Dict, List, Tuple
from mypy_extensions import TypedDict
from forte.utils import create_import_error_msg
from forte.models.srl import model_utils as utils
from forte.models.srl.data import SRLSpan, Span


try:
    import torch
    from torch import nn
    from torch.nn import functional as F
except ImportError as e:
    raise ImportError(
        create_import_error_msg("torch", "models", "Texar model support")
    ) from e


try:
    import texar.torch as tx
except ImportError as e:
    raise ImportError(
        create_import_error_msg(
            "texar-pytorch", "models", "Texar model support"
        )
    ) from e


class LabeledSpanGraphNetwork(tx.ModuleBase):
    @property
    def output_size(self):
        """
        This module is supposed to be the last layer so we will not return
        an informative output_size
        Returns:

        """
        return 0

    __torch_device__: torch.device

    def __init__(
        self, word_vocab: tx.data.Vocab, char_vocab: tx.data.Vocab, hparams=None
    ):
        super().__init__(hparams)

        # Word vocabulary & representation
        self.word_vocab = word_vocab
        self.word_embed = tx.modules.WordEmbedder(
            init_value=tx.data.Embedding(
                vocab=self.word_vocab.token_to_id_map_py,
                hparams={
                    "file": self._hparams.context_embeddings.path,
                    "dim": self._hparams.context_embeddings.size,
                    "read_fn": "load_glove",
                },
            ).word_vecs
        )
        self.head_embed = tx.modules.WordEmbedder(
            init_value=tx.data.Embedding(
                vocab=self.word_vocab.token_to_id_map_py,
                hparams={
                    "file": self._hparams.head_embeddings.path,
                    "dim": self._hparams.head_embeddings.size,
                    "read_fn": "load_glove",
                },
            ).word_vecs
        )
        self.span_length_embed = tx.modules.PositionEmbedder(
            position_size=self._hparams.max_arg_width,
            hparams={
                "dim": self._hparams.feature_size,
            },
        )

        # Character vocabulary & representation
        self.char_cnn = utils.CharCNN(
            char_vocab=char_vocab,
            hparams={
                "char_embed_size": self._hparams.char_embedding_size,
                "filter_widths": self._hparams.filter_widths,
                "filter_size": self._hparams.filter_size,
            },
        )
        self.embed_dropout = nn.Dropout(self._hparams.lexical_dropout_rate)

        # ELMo representation
        if self._hparams.elmo.path is not None:
            # pylint: disable=import-outside-toplevel
            from allennlp.modules.elmo import Elmo, batch_to_ids

            elmo_hparams = self._hparams.elmo
            self.elmo = Elmo(
                options_file=elmo_hparams.config,
                weight_file=elmo_hparams.path,
                num_output_representations=1,
            )
            self._elmo_char_ids_fn = batch_to_ids
        else:
            self.elmo = None

        # LSTM
        single_hidden_dim = self._hparams.contextualization_size
        lstm_input_dim = self.word_embed.dim + self.char_cnn.output_size
        if self.elmo is not None:
            lstm_input_dim += self._hparams.elmo.dim
        self.lstm = utils.CustomBiLSTM(
            hparams={
                "input_dim": lstm_input_dim,
                "hidden_dim": single_hidden_dim,
                "num_layers": self._hparams.contextualization_layers,
                "dropout": self._hparams.lstm_dropout_rate,
            }
        )
        hidden_dim = single_hidden_dim * 2

        self.label_vocab = {
            label: idx + 1  # reserve index 0 for null label
            for idx, label in enumerate(self._hparams.srl_labels)
        }
        self.label_inverse_vocab = {v: k for k, v in self.label_vocab.items()}
        self.head_attention = nn.Linear(hidden_dim, 1)

        word_input_dim = self.word_embed.dim + self.char_cnn.output_size
        mlp_num_layers = self._hparams.ffnn_depth
        mlp_hparams = {
            "input_sizes": [
                hidden_dim,  # concat'd state at start of span
                hidden_dim,  # concat'd state at end of span
                word_input_dim,
                self.span_length_embed.dim,
            ],
            "num_layers": mlp_num_layers,
            "hidden_size": [self._hparams.ffnn_size] * mlp_num_layers,
            "dropout_rate": self._hparams.dropout_rate,
        }
        self.argument_mlp = utils.ConcatInputMLP(
            hparams={
                **mlp_hparams,
                "output_size": 1,
                "activation": "ReLU",
            }
        )
        self.predicate_mlp = utils.ConcatInputMLP(
            hparams={
                **mlp_hparams,
                "input_sizes": [hidden_dim],
                "output_size": 1,
                "activation": "ReLU",
            }
        )
        self.span_label_mlp = utils.ConcatInputMLP(
            hparams={
                **mlp_hparams,
                "input_sizes": mlp_hparams["input_sizes"] + [hidden_dim],
                "output_size": len(self.label_vocab),
                "activation": "ReLU",
            }
        )

    @staticmethod
    def default_hparams() -> Dict[str, Any]:
        return {
            "filter_widths": [3, 4, 5],
            "filter_size": 50,
            "char_embedding_size": 8,
            "context_embeddings": {
                "path": "embeddings/glove.840B.300d.05.filtered",
                "size": 300,
                "datasets": "txt",
                "lowercase": False,
            },
            "head_embeddings": {
                "path": "embeddings/glove_50_300_2.filtered",
                # "path": "embeddings/glove_50_300_2.txt",
                "size": 300,
                "datasets": "txt",
                "lowercase": False,
            },
            "elmo": {
                "path": None,
                "config": None,
                "dim": 256,
            },
            "contextualizer": "lstm",
            "contextualization_size": 200,
            "contextualization_layers": 3,
            "ffnn_size": 150,
            "ffnn_depth": 2,
            "feature_size": 20,
            "max_span_width": 30,
            "model_heads": True,
            "num_attention_heads": 1,
            "srl_labels": [
                # predicate
                "V",
                # simple propositions
                "A0",
                "A1",
                "A2",
                "A3",
                "A4",
                "A5",
                "AA",
                "AM",
                "AM-ADV",
                "AM-CAU",
                "AM-DIR",
                "AM-DIS",
                "AM-EXT",
                "AM-LOC",
                "AM-MNR",
                "AM-MOD",
                "AM-NEG",
                "AM-PNC",
                "AM-PRD",
                "AM-REC",
                "AM-TM",
                "AM-TMP",
                # propositions with coreferenced arguments
                "C-A0",
                "C-A1",
                "C-A2",
                "C-A3",
                "C-A4",
                "C-A5",
                "C-AM-ADV",
                "C-AM-CAU",
                "C-AM-DIR",
                "C-AM-DIS",
                "C-AM-EXT",
                "C-AM-LOC",
                "C-AM-MNR",
                "C-AM-NEG",
                "C-AM-PNC",
                "C-AM-TMP",
                "C-V",
                # propositions with discontinuous argument
                "R-A0",
                "R-A1",
                "R-A2",
                "R-A3",
                "R-A4",
                "R-AA",
                "R-AM-ADV",
                "R-AM-CAU",
                "R-AM-DIR",
                "R-AM-EXT",
                "R-AM-LOC",
                "R-AM-MNR",
                "R-AM-PNC",
                "R-AM-TMP",
            ],
            "max_arg_width": 30,
            "argument_ratio": 0.8,
            "predicate_ratio": 0.4,
            "lexical_dropout_rate": 0.5,
            "dropout_rate": 0.2,
            "lstm_dropout_rate": 0.4,
        }

    @property
    def _device(self) -> torch.device:
        if not hasattr(self, "__torch_device__"):
            self.__torch_device__ = next(self.parameters()).device
        return self.__torch_device__

    def _create_span_indices(
        self, batch_size: int, max_len: int, max_span: int
    ) -> Tuple[torch.LongTensor, torch.LongTensor, torch.LongTensor]:
        start_ids = torch.arange(0, max_len).repeat_interleave(max_span)
        end_ids = start_ids + torch.arange(0, max_span).repeat(max_len)
        valid_mask = end_ids < max_len
        start_ids, end_ids = start_ids[valid_mask], end_ids[valid_mask]
        span_length = end_ids - start_ids
        start_ids = start_ids.expand(batch_size, *start_ids.size()).to(
            device=self._device
        )
        end_ids = end_ids.expand_as(start_ids).to(device=self._device)
        span_length = span_length.expand_as(start_ids).to(device=self._device)
        return start_ids, end_ids, span_length

    @staticmethod
    def _set_submatrix(mat: torch.Tensor, x: int, y: int, value: torch.Tensor):
        mat[x : (x + value.size(0)), y : (y + value.size(1))] = value

    def _create_softmax_mask(
        self, batch_size: int, max_len: int, max_span: int
    ) -> torch.ByteTensor:
        # 1 + 2 + ... + max_span + max_span + ...  (total max_len terms)
        total_lines = (1 + min(max_span, max_len)) * min(max_span, max_len) // 2
        if max_len > max_span:
            total_lines += (max_len - max_span) * max_span

        lower_tri = torch.tril(
            torch.ones(max_span, max_span, dtype=torch.uint8)
        )
        mask = torch.zeros(total_lines, max_len, dtype=torch.uint8)
        line_count = 0
        for idx in range(max_len):
            if max_len - idx < max_span:
                cur_mask = lower_tri[: (max_len - idx), : (max_len - idx)]
            else:
                cur_mask = lower_tri
            self._set_submatrix(mask, line_count, idx, cur_mask)
            line_count += cur_mask.size(0)
        mask = mask.expand(batch_size, total_lines, max_len)
        return mask.to(device=self._device)

    def _filter_labels(
        self,
        start_ids: torch.LongTensor,
        end_ids: torch.LongTensor,
        predicates: torch.LongTensor,
        srls: List[List[SRLSpan]],
    ) -> torch.LongTensor:
        batch_size, num_spans = start_ids.size()
        num_predicates = predicates.size(1)
        device = start_ids.device
        start_ids = start_ids.cpu().numpy()
        end_ids = end_ids.cpu().numpy()
        predicates = predicates.cpu().numpy()
        batch_predicates = [
            {pred: idx for idx, pred in enumerate(preds)}
            for preds in predicates
        ]
        batch_spans = [
            {(l, r): idx for idx, (l, r) in enumerate(zip(starts, ends))}
            for starts, ends in zip(start_ids, end_ids)
        ]

        gold_labels = torch.zeros(
            batch_size, num_predicates * num_spans, dtype=torch.long
        )
        for b_idx in range(batch_size):
            for srl in srls[b_idx]:
                span_idx = batch_spans[b_idx].get((srl.start, srl.end), None)
                predicate_idx = batch_predicates[b_idx].get(srl.predicate, None)
                if span_idx is not None and predicate_idx is not None:
                    label_idx = predicate_idx * num_spans + span_idx
                    gold_labels[b_idx, label_idx] = self.label_vocab[srl.label]
        gold_labels = gold_labels.to(device=device)
        return gold_labels

    def _compute_soft_head_attention_brute(
        self,
        start_ids: torch.LongTensor,
        end_ids: torch.LongTensor,
        sent_lengths: torch.LongTensor,
        states: torch.Tensor,
        word_inputs: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.LongTensor]:
        device = start_ids.device
        batch_size, max_len = states.size()[:2]
        num_spans = start_ids.size(1)
        max_span_width = self._hparams.max_span_width
        batch_offset = torch.arange(batch_size, device=device) * max_len
        span_indices = torch.arange(max_span_width, device=device)
        # span_indices: (batch_size, num_spans, max_span_width)
        span_indices = (
            span_indices.expand(batch_size, num_spans, -1)
            + start_ids.unsqueeze(-1)
            + batch_offset.view(-1, 1, 1)
        )
        # valid_spans: (batch_size, num_spans)
        valid_spans = end_ids < sent_lengths.unsqueeze(-1)
        # valid_spans_idx: (total_spans)
        valid_spans_idx = valid_spans.view(-1).nonzero().view(-1)
        # flat_span_indices: (total_spans, max_span_width)
        flat_span_indices = torch.index_select(
            span_indices.view(-1, max_span_width), dim=0, index=valid_spans_idx
        )

        # flat_sent_lengths: (total_spans)
        flat_sent_lengths = torch.index_select(
            (
                torch.min(end_ids + 1, sent_lengths.unsqueeze(-1))
                + batch_offset.unsqueeze(-1)
            ).view(-1),
            dim=0,
            index=valid_spans_idx,
        )
        # flat_mask: (total_spans, max_span_width)
        flat_mask = flat_span_indices < flat_sent_lengths.unsqueeze(-1)
        flat_span_indices *= flat_mask.type_as(flat_span_indices)

        # span_word_inputs: (total_spans, max_span_width, word_input_dim)
        span_word_inputs = torch.index_select(
            word_inputs.view(-1, word_inputs.size(-1)),
            dim=0,
            index=flat_span_indices.view(-1),
        ).view(*flat_span_indices.size(), -1)

        # logits: (batch_size, max_len)
        logits = self.head_attention(states).squeeze(-1)
        # flat_span_logits: (total_spans, max_span_width)
        flat_span_logits = torch.index_select(
            logits.view(-1), dim=0, index=flat_span_indices.view(-1)
        ).view(flat_span_indices.size())
        masked_span_logits = flat_span_logits - 1e10 * (~flat_mask).type_as(
            flat_span_logits
        )
        weights = torch.softmax(masked_span_logits, dim=-1)

        # weighted_inputs: (total_spans, max_span_width, word_input_dim)
        weighted_inputs = span_word_inputs * weights.unsqueeze(-1)
        # soft_head: (total_spans, word_input_dim)
        soft_head = torch.sum(weighted_inputs, dim=1)
        # indices: (batch_size, num_spans)
        indices = torch.cumsum(valid_spans.view(-1).type(torch.long), dim=0) - 1
        indices = torch.clamp_min(indices, 0).view_as(valid_spans)

        return soft_head, indices

    class ReturnType(TypedDict):
        loss: torch.Tensor
        total_scores: torch.Tensor
        start_ids: torch.LongTensor
        end_ids: torch.LongTensor
        predicates: torch.LongTensor

    def _arange(self, *args, **kwargs):
        return torch.arange(*args, device=self._device, **kwargs)

    def forward(self, inputs: tx.data.Batch) -> ReturnType:
        # Compute embeddings and recurrent states.
        char_embed = self.char_cnn(inputs.text)
        with torch.no_grad():
            # A workaround for freezing embeddings.
            word_embed = self.word_embed(inputs.text_ids)
            head_embed = self.head_embed(inputs.text_ids)
        context_embeds = [word_embed, char_embed]
        head_embeds = [head_embed, char_embed]
        if self.elmo is not None:
            char_ids = self._elmo_char_ids_fn(inputs.text).to(self._device)
            elmo_embed = self.elmo(char_ids)["elmo_representations"][0]
            context_embeds.append(elmo_embed)
        # *word_inputs: (batch_size, max_len, word_input_dim)
        lstm_word_inputs = self.embed_dropout(torch.cat(context_embeds, dim=-1))
        word_inputs = self.embed_dropout(torch.cat(head_embeds, dim=-1))

        # states: (batch_size, max_len, hidden_dim)
        states = self.lstm(lstm_word_inputs, inputs.length)

        # Create span indices.
        batch_size, max_len = inputs.text_ids.size()
        max_span = self._hparams.max_span_width
        # *_ids: (batch_size, max_num_spans)
        # max_num_spans ~= max_len * max_span
        start_ids, end_ids, span_length = self._create_span_indices(
            batch_size, max_len, max_span
        )
        # Create soft head representation weights.
        # head_attn_cache, head_attn_index = self._compute_soft_head_attention(
        (
            head_attn_cache,
            head_attn_index,
        ) = self._compute_soft_head_attention_brute(
            start_ids, end_ids, inputs.length, states, word_inputs
        )

        # Compute argument & predicate scores.
        span_length_embed = self.embed_dropout(self.span_length_embed.embedding)
        cache_inputs = [states, states, head_attn_cache, span_length_embed]
        pred_indices = self._arange(max_len).expand(batch_size, -1)
        with self.argument_mlp.cache_results(
            cache_inputs
        ), self.predicate_mlp.cache_results([states]):
            # arg_scores: (batch_size, max_num_spans)
            arg_scores = self.argument_mlp(
                [start_ids, end_ids, head_attn_index, span_length]
            ).squeeze(-1)
            # pred_scores: (batch_size, max_len)
            pred_scores = self.predicate_mlp([pred_indices]).squeeze(-1)

        # Beam pruning of arguments & predicates.
        # topk_*: (batch_size, max_arguments)
        max_arguments = math.ceil(self._hparams.argument_ratio * max_len)
        num_arguments = torch.ceil(
            self._hparams.argument_ratio * inputs.length.float()
        ).long()
        topk_arg_scores, topk_arg_indices = torch.topk(
            arg_scores, k=max_arguments, dim=1, sorted=True
        )
        topk_start_ids, topk_end_ids, topk_attn_index = utils.batch_gather(
            [start_ids, end_ids, head_attn_index], index=topk_arg_indices
        )
        topk_span_length = topk_end_ids - topk_start_ids

        # topk_pred_*: (batch_size, max_predicates)
        max_predicates = math.ceil(self._hparams.predicate_ratio * max_len)
        num_predicates = torch.ceil(
            self._hparams.predicate_ratio * inputs.length.float()
        ).long()
        topk_pred_scores, topk_pred_indices = torch.topk(
            pred_scores, k=max_predicates, dim=1, sorted=True
        )

        # Compute label scores for pruned argument-predicate pairs.
        with self.span_label_mlp.cache_results(cache_inputs + [states]):
            # label_scores:
            #   (batch_size, max_predicates * max_arguments, num_labels)
            label_scores = self.span_label_mlp(
                [
                    topk_start_ids.repeat(1, max_predicates),
                    topk_end_ids.repeat(1, max_predicates),
                    tx.utils.map_structure(
                        lambda x: x.repeat(1, max_predicates)
                        if isinstance(x, torch.Tensor)
                        else x,
                        topk_attn_index,
                    ),
                    topk_span_length.repeat(1, max_predicates),
                    topk_pred_indices.repeat_interleave(max_arguments, dim=1),
                ]
            )

        # Compute log-probabilities.
        total_scores = (
            topk_arg_scores.repeat(1, max_predicates).unsqueeze(-1)
            + topk_pred_scores.repeat_interleave(
                max_arguments, dim=1
            ).unsqueeze(-1)
            + label_scores
        )
        total_scores = torch.cat(
            [
                total_scores.new_zeros(*total_scores.size()[:-1], 1),
                total_scores,
            ],
            dim=-1,
        )
        gold_labels = self._filter_labels(
            topk_start_ids, topk_end_ids, topk_pred_indices, inputs.srl
        )

        # Compute masked loss.
        # unmasked_loss: (batch_size, max_predicates * max_arguments)
        unmasked_loss = F.cross_entropy(
            total_scores.view(-1, total_scores.size(-1)),
            gold_labels.view(-1),
            reduction="none",
        )
        # pred_*_mask: (batch_size, max_predicates)
        pred_index_mask = topk_pred_indices < inputs.length.unsqueeze(-1)
        pred_topk_mask = self._arange(max_predicates).unsqueeze(
            0
        ) < num_predicates.unsqueeze(1)
        # arg_*_mask: (batch_size, max_arguments)
        arg_index_mask = topk_end_ids < inputs.length.unsqueeze(-1)
        arg_topk_mask = self._arange(max_arguments).unsqueeze(
            0
        ) < num_arguments.unsqueeze(1)
        loss_mask = (
            (arg_index_mask & arg_topk_mask).unsqueeze(1)
            & (pred_index_mask & pred_topk_mask).unsqueeze(2)
        ).view(-1)
        loss = torch.sum(unmasked_loss * loss_mask.type_as(unmasked_loss))
        # loss = loss / batch_size
        return {
            "loss": loss,
            "total_scores": total_scores,
            "start_ids": topk_start_ids,
            "end_ids": topk_end_ids,
            "predicates": topk_pred_indices,
        }

    _CORE_ARGS = {
        f"{prefix}{arg}": 1 << idx
        for prefix in ["A", "ARG"]
        for idx, arg in enumerate("012345A")
    }

    def _dp_decode(
        self,
        max_len: int,
        pred_idx: int,
        start_ids: List[int],
        end_ids: List[int],
        argmax_labels: List[int],
        label_scores: List[float],
        enforce_constraint: bool = False,
    ) -> List[Span]:
        # Map positions to list of span indices for quick lookup during DP.
        spans_ending_at: Dict[int, List[int]] = defaultdict(list)
        for idx in range(  # pylint: disable=consider-using-enumerate
            len(end_ids)
        ):
            if argmax_labels[idx] == 0:  # ignore null spans
                continue
            if start_ids[idx] <= pred_idx <= end_ids[idx]:
                # Skip spans overlapping with the predicate.
                continue
            if end_ids[idx] >= max_len:
                # Skip invalid spans longer than the sentence.
                continue
            spans_ending_at[end_ids[idx]].append(idx)
        if all(len(spans) == 0 for spans in spans_ending_at.values()):
            return []  # no spans at all, just return
        if enforce_constraint:
            label_states = [
                self._CORE_ARGS.get(self.label_inverse_vocab[label], -1)
                if label != 0
                else -1
                for label in argmax_labels
            ]
        else:
            # ignore constraints
            label_states = [-1] * len(argmax_labels)

        # Perform DP.
        # Each state is a tuple (time, core_args), where `core_args` is the set
        # of core arguments (ARGA, ARG0 to ARG5) previously selected,
        # represented in binary (so {ARG0, ARG2, ARG5} would be
        # 2^0 + 2^2 + 2^5 = 37).
        max_scores = [{0: 0.0}]
        # only record selected spans
        best_span_indices: List[Dict[int, int]] = [{}]

        for idx in range(max_len):
            cur_scores = max_scores[-1].copy()
            cur_span_idx = {}
            for span_idx in spans_ending_at[idx]:
                label_state = label_states[span_idx]
                prev_states = max_scores[start_ids[span_idx]]
                for state, prev_score in prev_states.items():
                    if label_state != -1 and (label_state & state != 0):
                        # A core argument of this type has already been selected
                        continue
                    score = prev_score + label_scores[span_idx]
                    new_state = state | label_state
                    if score > cur_scores.get(new_state, 0):
                        cur_scores[new_state] = score
                        cur_span_idx[new_state] = span_idx
            max_scores.append(cur_scores)
            best_span_indices.append(cur_span_idx)

        # Backtrack to obtain optimal span choices.
        srl = []
        pos = max_len
        state = max(
            (score, state) for state, score in max_scores[max_len].items()
        )[1]
        while pos > 0:
            best_span_idx = best_span_indices[pos].get(state, None)
            if best_span_idx is not None:
                assert end_ids[best_span_idx] == pos - 1
                srl.append(
                    Span(
                        start_ids[best_span_idx],
                        end_ids[best_span_idx],
                        self.label_inverse_vocab[argmax_labels[best_span_idx]],
                    )
                )
                pos = start_ids[best_span_idx]
                if label_states[best_span_idx] != -1:
                    state &= ~label_states[best_span_idx]
            else:
                pos -= 1
        return srl

    @torch.no_grad()
    def decode(
        self, inputs: tx.data.Batch, enforce_constraint: bool = False
    ) -> List[Dict[int, List[Span]]]:
        r"""Performs optimal decoding with dynamic programming.

        :returns: A nested structure of SRL spans, representing the (inner) list
            of spans for each predicate (middle `dict`) and for each example in
            the batch (outer list).
        """
        result_dict = self.forward(inputs)
        start_ids = result_dict["start_ids"].cpu().numpy()
        end_ids = result_dict["end_ids"].cpu().numpy()
        predicates = result_dict["predicates"].cpu().numpy()
        batch_size, num_arguments = start_ids.shape
        num_predicates = predicates.shape[1]

        total_scores = result_dict["total_scores"].view(
            batch_size, num_predicates, num_arguments, -1
        )
        label_scores, argmax_label = torch.max(total_scores, dim=3)
        argmax_label = argmax_label.cpu().numpy()
        label_scores = label_scores.cpu().numpy()
        sent_lengths = inputs.length.cpu().numpy()

        # Do DP one example at a time...
        batch_srl = []
        for b_idx in range(batch_size):
            cur_srl: Dict[int, List[Span]] = {}
            # ... and one predicate at a time.
            for pred_idx, predicate in enumerate(predicates[b_idx]):
                if predicate >= inputs.length[b_idx]:
                    # Skip invalid predicates outside the sentence.
                    continue
                srl = self._dp_decode(
                    sent_lengths[b_idx],
                    predicate,
                    start_ids[b_idx],
                    end_ids[b_idx],
                    argmax_label[b_idx, pred_idx],
                    label_scores[b_idx, pred_idx],
                    enforce_constraint,
                )
                if len(srl) > 0:
                    cur_srl[predicate] = srl
            batch_srl.append(cur_srl)
        return batch_srl



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
    Exception classes for ontology code generation.
"""


class OntologyGenerationWarning(UserWarning):
    """General warning for ontology generation."""

    pass


class DirectoryAlreadyPresentWarning(OntologyGenerationWarning):
    """The directory that the code will be written to already exists."""

    pass


class DuplicateEntriesWarning(OntologyGenerationWarning):
    """The entry is defined multiple times."""

    pass


class DuplicatedAttributesWarning(OntologyGenerationWarning):
    """The attribute is defined multiple times."""

    pass


class OntologySpecError(ValueError):
    """General error related to ontology specification."""

    pass


class OntologySpecValidationError(OntologySpecError):
    """Error thrown during validating the ontology specification."""

    pass


class OntologySourceNotFoundException(OntologySpecError):
    """Raise when the import source specification cannot be found."""

    pass


class OntologyAlreadyGeneratedException(OntologySpecError):
    """Raise when generating the same ontology again, which is considered
    to be a cyclic or duplicated declaration."""

    pass


class ParentEntryNotDeclaredException(OntologySpecError):
    """Raise when the entry's parent entry is not previously defined."""

    pass


class ParentEntryNotSupportedException(OntologySpecError):
    """Raise when the entry uses a not allowed parent entry."""

    pass


class TypeNotDeclaredException(OntologySpecError):
    """
    Raise when the entry uses an attribute which is previously not defined.
    """

    pass


class UnsupportedTypeException(OntologySpecError):
    """Raise when the ontology generator do not support such type structure."""

    pass


class InvalidIdentifierException(OntologySpecValidationError):
    """Raise when the ontology contains illegal identifier."""

    pass


class CodeGenerationException(BaseException):
    """General errors during ontology generation."""

    pass



# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import random
import json
from typing import Tuple, Union, Dict, Any


from forte.data.ontology import Annotation
from forte.processors.data_augment.algorithms.single_annotation_op import (
    SingleAnnotationAugmentOp,
)
from forte.common.configuration import Config
from forte.utils import create_import_error_msg

__all__ = [
    "UniformTypoGenerator",
    "TypoReplacementOp",
]


class UniformTypoGenerator:
    r"""
    A uniform generator that generates a typo from a typo dictionary.

    Args:
        word: input word that needs to be replaced,
        dict_path: the url or the path to the pre-defined typo json file.
            The key is a word we want to replace. The value is a list
            containing various typos of the corresponding key.

            .. code-block:: python

                {
                    "apparent": ["aparent", "apparant"],
                    "bankruptcy": ["bankrupcy", "banruptcy"],
                    "barbecue": ["barbeque"]
                }
    """

    def __init__(self, dict_path: str):
        try:
            import requests  # pylint: disable=import-outside-toplevel
        except ImportError as e:
            raise ImportError(
                create_import_error_msg(
                    "requests", "data_aug", "data augment support"
                )
            ) from e

        try:
            r = requests.get(dict_path)
            self.data = r.json()
        except requests.exceptions.RequestException:
            with open(dict_path, encoding="utf8") as json_file:
                self.data = json.load(json_file)

    def generate(self, word: str) -> str:
        if word in self.data.keys():
            result: str = random.choice(self.data[word])
            return result
        else:
            return word


class TypoReplacementOp(SingleAnnotationAugmentOp):
    r"""
    This class is a replacement op using a pre-defined
    spelling mistake dictionary to simulate spelling mistake.
    """

    def __init__(self, configs: Union[Config, Dict[str, Any]]):
        super().__init__(configs)
        if "dict_path" in configs.keys():
            self.dict_path = configs["dict_path"]
        else:
            # default typo dictionary
            self.dict_path = (
                "https://raw.githubusercontent.com/wanglec/"
                + "temporaryJson/main/misspelling.json"
            )
        if configs["typo_generator"] == "uniform":
            self.typo_generator = UniformTypoGenerator(self.dict_path)
        else:
            raise ValueError(
                "The valid options for typo_generator are [uniform]"
            )

    def single_annotation_augment(
        self, input_anno: Annotation
    ) -> Tuple[bool, str]:
        r"""
        This function replaces a word from a typo dictionary.

        Args:
            input_anno: The input annotation.
        Returns:
            A tuple, where the first element is a boolean value indicating
            whether the replacement happens, and the second element is the
            replaced string.
        """
        # If the replacement does not happen, return False.
        if random.random() > self.configs.prob:
            return False, input_anno.text
        word: str = self.typo_generator.generate(input_anno.text)
        return True, word

    @classmethod
    def default_configs(cls):
        r"""
        Returns:
            A dictionary with the default config for this processor.
        Following are the keys for this dictionary:
            - prob (float): The probability of replacement,
              should fall in [0, 1]. Default value is 0.1
            - dict_path (str): the `url` or the path to the pre-defined
              typo json file. The key is a word we want to replace.
              The value is a list containing various typos
              of the corresponding key.
            - typo_generator (str): A generator that takes in a word and
              outputs the replacement typo.
        """
        return {
            "prob": 0.1,
            "dict_path": "https://raw.githubusercontent.com/wanglec/"
            + "temporaryJson/main/misspelling.json",
            "typo_generator": "uniform",
        }
